// ################################################################################
//
//  16 Ch Controller NodeMCU v0.00
//
//  Released:  18/07/2020
//
//  Author: TechKnowTone
//
// ################################################################################
/*
    TERMS OF USE: This software is furnished "as is", without technical support, and
    with no warranty, expressed or implied, as to its usefulness for any purpose. In
    no event shall the author or copyright holder be liable for any claim, damages,
    or other liability, whether in an action of contract, tort or otherwise, arising
    from, out of or in connection with the software or the use or other dealings in
    the software.

    microcontroller:  ESP8266 NodeMCU

    This code works in conjunction with a 'Processing' application to enable the user
    to control up to 16 servo motors connected to a PCA9685 controller. It responds
    to simple commands received on its serial port.

    This version was modified for use with RoboDog, to take advantage of the OLED
    display and battery monitoring system.

    The following commands are supported:
    !       - RESET and export default values to 16-Ch controller app
    SFnn.   - set PWM frequency, nominally 50 Hz
    SLnn.   - set 'target' lower limit to nn
    SOn.    - set servo board OE. 0 = LOW, 1 = HIGH
    STnn.   - set 'target' servo No. to nn
    SUnn.   - set 'target' upper limit to nn
    SVnn.   - set 'target' servo PWM value to nn
    QM.     - respond with MQ2. for WiFi mode 2
*/
// Declare Libraries
#include <Wire.h>
#include <Adafruit_PWMServoDriver.h>
#include <Adafruit_NeoPixel.h>  // only included to turn OFF random pixels on power-up
#include "SSD1306Wire.h"        // OLED display library

// create I2C PWM driver instance for PCA9685 board
Adafruit_PWMServoDriver I2Cpwm = Adafruit_PWMServoDriver(0x40);

// Define ESP8266 GPIO pin mapping
#define D0  16            // map pin to GPIO
#define D1   5            // map pin to GPIO
#define D2   4            // map pin to GPIO
#define D3   0            // map pin to GPIO
#define D4   2            // map pin to GPIO
#define D5  14            // map pin to GPIO
#define D6  12            // map pin to GPIO
#define D7  13            // map pin to GPIO
#define D8  15            // map pin to GPIO
#define D9   3            // map pin to GPIO
#define D10  1            // map pin to GPIO

// Define general constants
#define BatCal 103.66     // calibrated voltage multiplying factor, default 102.3 
#define BatCritical 726   // critical battery threshold, default = 700, set == 0 to ignore
#define BatMax 850        // A0 for fully charged battery voltage, == 8.2v
#define BattPin A0        // analogue pin used for battery monitoring
#define BatUSB 100        // if <= then USB supply assumed
#define ledPin0 D7        // RGB LED serial pin for mouth LEDs
#define ledPin1 D8        // RGB LED serial pin for eye LEDs
#define ledNum0 5         // number of body RGB LEDs connected
#define ledNum1 2         // number of eye RGB LEDs connected
#define Pin_OE D6         // define PCA9685 OE pin as D6

// NeoPixels code only included to reset pixels on power up
//Adafruit_NeoPixel strip = Adafruit_NeoPixel(ledNum, ledPin, NEO_GRB + NEO_KHZ800);
Adafruit_NeoPixel strip0 = Adafruit_NeoPixel(ledNum0, ledPin0, NEO_GRB + NEO_KHZ800);
Adafruit_NeoPixel strip1 = Adafruit_NeoPixel(ledNum1, ledPin1, NEO_GRB + NEO_KHZ800);

// Define PWM freq and servo calibration arrays
#define Freq 50  // MG996R servos run at ~50 Hz updates
// Servo:   { 0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F }
int SLL[] = {102,126,261,125,344,143,229,100,100,100,160,357,122,268,135,124};  // servo lower limit values
int SRV[] = {260,325,460,125,394,189,332,300,300,300,209,404,122,472,344,258};  // reference values used to set a servo
int SUL[] = {463,478,460,332,440,230,424,500,500,500,251,446,329,472,502,467};  // servo upper limit values

// Label:   {FL1 FR1 FL2 FR2 FL3 FR3 SH0 NC  NC  NC  RR3 RL3 RR2 RL2 RR1 RL1}


// Display constants
const int I2C_DISPLAY_ADDRESS = 0x3c;
const int SDA_PIN = D2;
const int SDC_PIN = D1;

// Initialize the oled display for address 0x3c
 SSD1306Wire display(I2C_DISPLAY_ADDRESS, SDA_PIN, SDC_PIN);

// Declare and initialise global variables
long AnyLong;                 // any temp value
int BatAvg;               // average battery voltage
int BatSum;              // cumulative battery voltage
unsigned long BatT0 = 0;      // time of first battery measurement in life prediction forecast
String BatTime$ = "--:--";    // time left displayed
unsigned long BatTms;         // time when battery reading was taken
int BatV0 = 0;                // reference voltage used in prediction forecast
int BatV1 = 0;                // lowest voltage used in prediction forecast
int BattVol;              // instantaneous battery voltage
bool Cal = false;         // if == true enables immediate rest, only set after calibration
char cmdMode = ' ';       // command mode
char cmdType = ' ';       // command mode type
int cmdVal = 0;           // value associated with a cmdType
bool DispClr = false;     // set == true to clear the memory contents
int DispCnt = 50;         // counter used in display updating
bool DispDisp = false;    // set = true to immediately display memory contents
char keyChar;             // any keyboard character
int keyVal;               // any keyboard value
int mouthCnt = 0;         // counter used in mouth LED tasks
unsigned long next20ms;   // 20ms loop timer
unsigned long next40ms;   // 40ms print loop timer
String PrintTx = "";      // printed strings
float pwmFreq = Freq;     // MG996R servos run at ~50 Hz updates
byte SerialRx = 0;        // serial receive flag; 0 = Serial, 1 = Serial1
byte servoOE = 1;         // servo board Output Enable; default = 1  OFF
int Servo_PWM;            // servo pulse width 800 - 2200
int servoTgt = 0;         // target servo for manual setting; default = 0

void setup() {
  // Setup code, runs once:
  // Initialise 16 x 12-bit PWM driver I2C interface immediately
  pinMode(Pin_OE, OUTPUT);
  digitalWrite(Pin_OE, HIGH); // disable all PWM outputs to start with

  pinMode(BattPin,INPUT); BattVol = analogRead(BattPin);  // capture the battery voltage

  Serial.begin(115200);   // baud rate for Rx/Tx comms

  // Initialise neopixel RGB LEDs
  strip0.begin();
  strip0.fill(0,0,5);  // strip.fill(color, first, count);
  strip0.show();       // Initialize all pixels to 'off'
  strip1.begin();
  strip1.fill(0,0,2);  // strip.fill(color, first, count);
  strip1.show();       // Initialize all pixels to 'off'
  delay(10);
  NeoPixel_SetAllMthLEDs(0,0,0);

  Wire.begin();
  I2Cpwm.begin();
  I2Cpwm.setPWMFreq(pwmFreq);  // set PCA9685 PWM frequency
  if (Cal) {GoToRest();}      // go immediately to rest position
  ClearPCAop();

  PrintTx = "\n\n16-Ch Controller NodeMCU v0.00\n";
  SetDefaults();

  Display_Init();
  Display_Intro();
  
  next20ms = millis() + 10; // reset 20ms loop timer
  next40ms = millis() + 20; // reset 40ms loop timer with 10ms phase to 20ms timer
  ExportArrays();           // send the contents of the arrays to the 16-Ch controller app
}

// ---------------------------------------------------------------------------------

void(* resetFunc) (void) = 0;//declare reset function at address 0

// ---------------------------------------------------------------------------------

void loop() {
  // main code which runs repeatedly as multi-tasking functions
  keyVal = Serial.read();
  if (keyVal != -1) {SerialRx = 0; decodeKey(keyVal); DispCnt = 50;}

  ////////////////////////////////////////////////////////////////////////////////
  //
  //  20 ms tasks
  //
  ////////////////////////////////////////////////////////////////////////////////
  // Every 20ms we read the battery voltage and set a lamp periodically to indicate
  // is capacity
  if (millis() >= next20ms) {
    next20ms = millis() + 20; // set next target time
    readBattery();
  }

  ////////////////////////////////////////////////////////////////////////////////
  //
  //  40 ms Print subtask
  //
  ////////////////////////////////////////////////////////////////////////////////
  // Every 40ms we check to see if there is anything in the print buffer and if so
  // we send a portion (32 chars) of it.
  if (millis() >= next40ms) {
    next40ms = millis() + 40; // set next target time
    PrintTxPacket();
    Display_40ms();
  }
}

// ---------------------------------------------------------------------------------

void SetDefaults() {
  // load a range of default values  
  BatAvg = 1023;           // average battery voltage
  BatSum = 20460; //  pre-set cumulative battery voltage
}
