
// ---------------------------------------------------------------------------------

void clearMem() {
  // Clear the contents of the move memory Mem[] before loading values
  int zMax = MemMax * MemWidth;
  for (int zM = 0; zM < zMax; zM++) {Mem[zM] = 0;}

  // Clear flags ready for a new move
  MemRow = 0; MemMax = 0; Mem1st = 0;
  CmdFlag = 0; setBalanceOFF(); CmdForNum = 0; CmdForRow = 0;
}

// ---------------------------------------------------------------------------------

void clearMoveFlags() {
  // called before setting a flag to ensure that the correct ones are then set
  AtRest = false;
  Ready = false;
  SitHigh = false;
  SitLow = false;
  Stand = false;
}

// ---------------------------------------------------------------------------------

void getHeadAngle() {
  // converts the present head servo PWM to an angle
  HeadAngle = map(SPP[SH0],SLL[SH0],SUL[SH0],45,135);
}

// ---------------------------------------------------------------------------------

void getSpeedJXY(int zJXY) {
  // Joystick vector JoyV varies from 12 - 127 which is mapped into a speed profile
  // low to medium demands varies speed between SpdMin and SpdTop, where SpdTop = SpdMid
  // higher demands increase SpdTop up to SpdMax, extending the SpdTop speed range
  // reduced demand lowers SpdTop towards SpdMid
  // Gear changes from 1 - 5 to represent speed from SpdMin - SpdMax
  if (zJXY >= 120) {
    // high demand so increase speed progressively
     SpeedCnt++; if (SpeedCnt >= 6) {SpeedCnt = 0; SpdTop++;}
  } else if (zJXY <= 40) {
    // low demand so decrease speed progressively
     SpeedCnt++; if (SpeedCnt >= 3) {SpeedCnt = 0; SpdTop--;}
  } else {SpeedCnt = 0;}
  SpdTop = min(SpdTop,SpdMax); // limit maximum top speed
  SpdTop = max(SpdTop,SpdMid); // limit minimum top speed
  
  // limit rate of change of speed to avoid surges in demand
  SpeedTgt = map(zJXY,12,127,SpdMin,SpdTop);
  if (Speed < SpeedTgt) {Speed += 1 + ((SpeedTgt - Speed)/4);}
  else if (Speed > SpeedTgt) {Speed -= 1 + ((Speed - SpeedTgt)/4);}
}

// ---------------------------------------------------------------------------------

void getSteps() {
  // scans the target values for the largest difference and returns it in steps
  Steps = 4; int zD;
  for (int zS = 0; zS < 16; zS++) {
    zD = abs(STG[zS] - SPP[zS]); if (zD > Steps) {Steps = zD;} 
  }
//  PrintTx += "\nSteps = " + String(Steps) + "\n";
}

// ---------------------------------------------------------------------------------

void getServoAngle(int zCh) {
  // Converts the servo channel zCh PWM to an angle value in degrees
  int zA0 = 30; int zA1 = 150;
  switch(zCh) {
      case SFL1: zA0 = 160; zA1 =   0; break;
      case SFL2: zA0 =  90; zA1 =   0; break;
      case SFL3: zA0 =  50; zA1 = 110; break;

      case SFR1: zA0 =   0; zA1 = 160; break;
      case SFR2: zA0 =   0; zA1 =  90; break;
      case SFR3: zA0 =  70; zA1 = 130; break;

      case SRL1: zA0 = 160; zA1 =   0; break;
      case SRL2: zA0 =  90; zA1 =   0; break;
      case SRL3: zA0 =  50; zA1 = 110; break;

      case SRR1: zA0 =   0; zA1 = 160; break;
      case SRR2: zA0 =   0; zA1 =  90; break;
      case SRR3: zA0 =  70; zA1 = 130; break;
  }

  // non-linearity of servo is corrected
  if (SPP[zCh] >= SRV[zCh]) {Angle = map(SPP[zCh],SRV[zCh],SUL[zCh],90,zA1);}
  else {Angle = map(SPP[zCh],SLL[zCh],SRV[zCh],zA0,90);}
  
//  Angle = map(SPP[zCh],SLL[zCh],SUL[zCh],zA0,zA1);
}

// ---------------------------------------------------------------------------------

void GoToRest() {
  // called at start when servo positions are unknown
  // immediately drives servos to default PWM rest positions
  // waits one second then removes pulse and OE
  for (int zS = 0; zS < 16; zS++) {
    I2Cpwm.setPWM(zS, 0, SRV[zS]);
  }
  digitalWrite(Pin_OE, LOW); // enable all PWM outputs
  AtRest = true; AutoOffTimer = 200; // set 2s timeout
}

// ---------------------------------------------------------------------------------

void Head_Engine() {
  // head movement uses a separate timer so that the speed of movement can be set
  // independantly  
  if (HeadTOF) {
    // timer overflow has occured so wait for micros() timer to overflow too
    if (micros() < lastHeadMicros) {HeadTOF = false;}  // back to normal, clear the flag
    else {return;}  // return to do other things in the mean time
  }
  headMicros = micros() + headInterval; // set next time point
  lastHeadMicros = micros(); if (headMicros < micros()) {HeadTOF = true;} // time point has over-flowed, so set overflow flag
  if (Pause) {return;}  // pause flag causes movement to be put on hold
  if (Head > 0) {
    if (servoOE) {SetServoOE(0);} // turn the servos ON
    if (HeadStep > 1) {
      SPP[SH0] = STG[SH0] - (((STG[SH0] - SSP[SH0])*HeadStep)/HeadSteps);
      I2Cpwm.setPWM(SH0, 0, SPP[SH0]);
      HeadStep--;
    } else if (HeadStep == 1) {
      SPP[SH0] = STG[SH0]; SSP[SH0] = STG[SH0];
      I2Cpwm.setPWM(SH0, 0, SPP[SH0]);
      HeadStep = 0;
    } else {
      Head = 0;
    }
  }
}

// ---------------------------------------------------------------------------------

// Order: SFR1,SFR2,SFR3,SFL1,SFL2,SFL3,SRR1,SRR2,SRR3,SRL1,SRL2,SRL3
void MemSet(int zFR1,int zFR2,int zFR3,int zFL1,int zFL2,int zFL3,int zRR1,int zRR2,int zRR3,int zRL1,int zRL2,int zRL3,int zSt) {
  // function is called to load a set of angle values into the move array
  // the value of MemRow must be set before calling this function, normally = 0
  // at the end MemRow will be incremented to point at the next row so subsequent
  // calls will load the memory sequentially
  int zP = MemRow * MemWidth; MemRow ++; MemMax = MemRow;
  if ((!Trans) || (zFR1 < 0)) {
    // load in angles normally
    // this is ALWAYS true for command rows which must not be transposed
    Mem[zP] = zFR1; zP++;
    Mem[zP] = zFR2; zP++;
    Mem[zP] = zFR3; zP++;

    Mem[zP] = zFL1; zP++;
    Mem[zP] = zFL2; zP++;
    Mem[zP] = zFL3; zP++;

    Mem[zP] = zRR1; zP++;
    Mem[zP] = zRR2; zP++;
    Mem[zP] = zRR3; zP++;

    Mem[zP] = zRL1; zP++;
    Mem[zP] = zRL2; zP++;
    Mem[zP] = zRL3; zP++;
  } else {
    // transpose the loading of angles
    // it is the level 3 angles which move in opposition, FR3, FL3, RR3, RL3
    // note these angles can have Pwr and Rate flags attached, which need to be
    // transposed too
    Mem[zP] = zFL1; zP++;
    Mem[zP] = zFL2; zP++;
    Mem[zP] = 180 - getAngVal(zFL3);
    if (AngRate) {Mem[zP] += 400;}
    if (!AngPwr) {Mem[zP] += 200;}
    zP++;

    Mem[zP] = zFR1; zP++;
    Mem[zP] = zFR2; zP++;
    Mem[zP] = 180 - getAngVal(zFR3);
    if (AngRate) {Mem[zP] += 400;}
    if (!AngPwr) {Mem[zP] += 200;}
    zP++;

    Mem[zP] = zRL1; zP++;
    Mem[zP] = zRL2; zP++;
    Mem[zP] = 180 - getAngVal(zRL3);
    if (AngRate) {Mem[zP] += 400;}
    if (!AngPwr) {Mem[zP] += 200;}
    zP++;

    Mem[zP] = zRR1; zP++;
    Mem[zP] = zRR2; zP++;
    Mem[zP] = 180 - getAngVal(zRR3);
    if (AngRate) {Mem[zP] += 400;}
    if (!AngPwr) {Mem[zP] += 200;}
    zP++;
  }
  Mem[zP] = zSt;
}

// ---------------------------------------------------------------------------------

void Move_As_Memory(int zRow, byte zMode,int zSpd) {
  // loads target angles stored in Mem[] as a move sequence.
  // zMode == 0 - do this once
  // zMode == 1 - do this continuously
  // this function sets up the move engine to load data automatically
  // the 'Move' calling function must monitor the move
  // zRow is normally set by Mem1st as the 1st row to be played by the Walk engine
  if (MoveMemTgtLoader(zRow)) {
    // load memory row 0 as angle targets and speed
    WalkMode = zMode;           // set mode as once == 0 or continuous == 1
    Speed = zSpd;
    if (Quick) {Speed = 100; Quick = false;}  // speed over-ride flag
    Walk_Begin();               // initiate walking process
  }
}

// ---------------------------------------------------------------------------------

void MoveBkd() {
  // called from move state machine
  if (MoveState != 5) {
    // just been asked to move backwards
    MoveState = 5; LEDMthTask = 20;
    if (AtRest) {
      // at reast so we need to move to ready position
      MoveToReady(1); loopWhileWalk();
    }
    // now initiate walking
    if (Ready) {MoveToWalkBackwards();}
    else {MoveState = -1; return;}
  }
  if (JoyD == 5) {
    // demand is still there so keep moving at demanded speed
    getSpeedJXY(JoyV);
    setLegSpeed(Speed);
  } else {
    // -Y demand has dropped so return to ready state
    MoveState = -1;
  }
}

// ---------------------------------------------------------------------------------

void MoveFwd() {
  // called from move state machine
//  Serial.println("MoveFwd()");
  if (MoveState != 1) {
    // just been asked to move forward
    MoveState = 1; LEDMthTask = 20;
    if (Stand) {
      // stood up straight so we need to move to ready position
      MoveToReady(1); loopWhileWalk();
    }
    // now initiate walking
    if (Ready) {MoveToWalk();}
    else {MoveState = -1; return;}
  }
  if (JoyD == 1) {
    // demand is still there so keep moving at demanded speed
    getSpeedJXY(JoyV);
    setLegSpeed(Speed);
  } else {
    // Y demand has dropped so return to ready state
    MoveState = -1;
  }
}

// ---------------------------------------------------------------------------------

void MoveFwdLft() {
  // move forward with a bias twist towards the left  
  if (MoveState != 8) {
    // just been asked to move forward towards the left
    MoveState = 8; LEDMthTask = 20;
    if (AtRest) {
      // at reast so we need to move to ready position
      MoveToReady(1); loopWhileWalk();
    }
    if (Gear <= 2) {
//      MoveToSkiFwdLft();
    } else {
//      MoveToWalkFwdLft();
    } 
  }
  if (JoyD == 8) {
    // demand is still there so keep moving at demanded speed
    getSpeedJXY(JoyV);
    setLegSpeed(Speed);
  } else {
    // Y demand has dropped so return to ready state
    MoveState = -1;
  }
}

// ---------------------------------------------------------------------------------

void MoveFwdRht() {
  // move forward with a bias twist towards the right  
  if (MoveState != 2) {
    // just been asked to move forward towards the left
    MoveState = 2; LEDMthTask = 20;
    if (AtRest) {
      // at reast so we need to move to ready position
      MoveToReady(1); loopWhileWalk();
    }
    if (Gear <= 2) {
//      MoveToSkiFwdRht();
    } else {
//      MoveToWalkFwdRht();
    }
  }
  if (JoyD == 2) {
    // demand is still there so keep moving at demanded speed
    getSpeedJXY(JoyV);
    setLegSpeed(Speed);
  } else {
    // Y demand has dropped so return to ready state
    MoveState = -1;
  }
}

// ---------------------------------------------------------------------------------

void MoveHeadTo(int zAngle, int zSpeed) {
  // move the head to specific angle  
  SetStartsToPPs();
  setHeadTgt(zAngle); HeadSteps = abs(STG[SH0] - SPP[SH0]); HeadStep = HeadSteps;
  setHeadSpeed(zSpeed); Head = 1;
}

// ---------------------------------------------------------------------------------

bool MoveMemTgtLoader(int zRow) {
  // load Mem[zRow+] angle target values until a row with all zeros is encountered
  // or zR >= MemMax, at which point we reset to zero
  // This function also detects the following embedded commands:
  //  0 - 'Blank' line, ignore
  //  1 - goto a specified row, branch instruction used to loop the move
  //  2 - set the speed to a new value
  //  3 - set a pause period in steps
  //  4 - set the CmdFlag to a value
  //  5 - turn self-balance ON/OFF 
  //  6 - For... 1 to n loop
  //  7 - Next in a For... loop
  Start:
  int zRowRef = zRow; // remember row value passed to this function
  // only report memory load if connected to the MoveTrainer
  if (GR == 0) {PrintTx += "MT " + String(zRow) + "\n";}
  bool zVal = false;  // flag used to determine if any targets are set
  WalkMem = zRow + 1; // set the pointer for next time
  zRow *= MemWidth;   // point to first value in zRow
  
  // look for command lines, with first value < 0
  if (Mem[zRow] < 0) {
    // this is a command line, so decode it
    int zCmd = Mem[zRow + 1]; // get the command type
    // this is a command row, so decode it
    if (zCmd == 0) {
      // Blank line, so skip over it
      if (MTM_Step) {return zVal;}  // single stepping so don't go to next row
      if ((zRowRef + 1) >= MemMax) {return zVal;} // we are at the last row
      zRow = zRowRef + 1; goto Start;
    }
    else if (zCmd == 1) {
      // this is a goto command, so load revised row pointer and loop back
      if (WalkMode < 1) {return zVal;}  // don't branch if WalkMode == 0 one shot mode
      zRow = Mem[zRow + 2]; moveCycleCnt++; goto Start;
    }
    else if (zCmd == 2) {
      // this is a speed command, so load revised speed and loop back for next row
      Speed = Mem[zRow + 2]; // set revised speed
//      PrintTx += "Speed " + String(Speed) + "\n";
      if (MTM_Step) {return zVal;}  // single stepping so don't go to next row
      if ((zRowRef + 1) >= MemMax) {return zVal;} // we are at the last row
      zRow = zRowRef + 1; goto Start;
    }
    else if (zCmd == 3) {
      // this is a pause command, so load time delay only and perform single step
      // to where we are now
      if (MTM_Step) {return zVal;}  // single stepping so don't go to next row
      if ((zRowRef + 1) >= MemMax) {return zVal;} // we are at the last row
      // set things up for a multiple step to current position
      // as the servo present positions match their targets, PWM values will no change
      zRow += 2; // point at pause value
      LegSteps = Mem[zRow]; LegStep = LegSteps; RateProf = 0; zVal = true;
    }
    else if (zCmd == 4) {
      // this is a Flag command, so load the flag value, used by observing functions
      CmdFlag = Mem[zRow + 2];      // grab Flag value
      if (MTM_Step) {return zVal;}  // single stepping so don't go to next row
      if ((zRowRef + 1) >= MemMax) {return zVal;} // we are at the last row
      zRow = zRowRef + 1; goto Start;
    }
    else if (zCmd == 5) {
      // this is a Balance command, so load the value and turn balance ON/OFF
      if (Mem[zRow + 2] == 0) {
        setBalanceOFF();  // turn self-balance OFF
      } else {
        setBalanceON(Mem[zRow + 2]);   // turn self-balance ON
      }
      if (MTM_Step) {return zVal;}  // single stepping so don't go to next row
      if ((zRowRef + 1) >= MemMax) {return zVal;} // we are at the last row
      zRow = zRowRef + 1; goto Start;
    }
    else if (zCmd == 6) {
      // this is a For..., so load pointers tof the loop, then move on
      CmdForNum = Mem[zRow + 2];  // grab loops value
      CmdForRow = zRowRef + 1;    // store next row as NEXT  pointer
      if (MTM_Step) {return zVal;}  // single stepping so don't go to next row
      if ((zRowRef + 1) >= MemMax) {return zVal;} // we are at the last row
      zRow = zRowRef + 1; goto Start;
    }
    else if (zCmd == 7) {
      // this is a Next in a For... loop, so decrement loop and jump to the stored row
      if (CmdForNum > 1) {
        // still some loops to do
        CmdForNum--; zRow = CmdForRow;
      } else {zRow = zRowRef + 1;}
      if (MTM_Step) {return zVal;}  // single stepping so don't go to next row
      goto Start;
    }
  } else {
    // this is not a command line, so deal with it as angles
    WalkCnt += WalkInc; // adjust the walk movement counter
    
    // load angles into servo targets and set power mask
    for (int zP = 0; zP < 12; zP++) {
      int zA = Mem[zRow + zP];
      if (zA >= 400) {zA -= 400; RateMask[zP] = true;} else {RateMask[zP] = false;} // set and remove the auto-rate flag
      if (zA >= 200) {PwrMask[zP] = false;} else {PwrMask[zP] = true;}  // set the power ON mask
      setServoTgt(zA,ChMap[zP]);
    } zVal = true;
  
    // separate steps from rate profile
    // you must set the speed after loading a row as steps and profile can change
    zRow += 12; 
    LegSteps = Mem[zRow]; RateProf = 0;           // constant speed
  //  PrintTx += "LegSteps " + String(LegSteps) + "\n";
         if (LegSteps > 600) {LegSteps -= 600; RateProf = 3;}  // acceleration then deceleration
    else if (LegSteps > 400) {LegSteps -= 400; RateProf = 2;}  // deceleration
    else if (LegSteps > 200) {LegSteps -= 200; RateProf = 1;}  // acceleration
    LegStep = LegSteps;
  //  PrintTx += "RateProf " + String(RateProf) + "\n";
    
    if (Mem[zRow] > 0) {zVal = true;}  // steps set for this row so flag as ok
  }
  return zVal;
}

// ---------------------------------------------------------------------------------

void MoveStepLeft() {
  // move by stepping to the left
  if (MoveState != 17) {
    // just been asked to step left
    MoveState = 17; LEDMthTask = 20;
    if (AtRest) {
      // at reast so we need to move to ready position
      MoveToReady(1); loopWhileWalk();
    }
    if (Ready) {MoveToStepLeft();}
    else {MoveState = -1; return;}
  }
  if (JoyD == 7) {
    getSpeedJXY(JoyV);
    setLegSpeed(Speed);
//    Serial.println(Steer);
  } else {
    // Y demand has dropped so return to ready state
    MoveState = -1;
  }
}

// ---------------------------------------------------------------------------------

void MoveStepRight() {
  // move by stepping to the right
  if (MoveState != 13) {
    // just been asked to move
    MoveState = 13; LEDMthTask = 20;
    if (AtRest) {
      // at reast so we need to move to ready position
      MoveToReady(1); loopWhileWalk();
    }
    if (Ready) {MoveToStepRight();}
    else {MoveState = -1; return;}
  }
  if (JoyD == 3) {
    getSpeedJXY(JoyV);
    setLegSpeed(Speed);
//    Serial.println(Steer);
  } else {
    // Y demand has dropped so return to ready state
    MoveState = -1;
  }
}

// ---------------------------------------------------------------------------------

void MoveTgtLoader(int zFR1,int zFR2,int zFR3,int zFL1,int zFL2,int zFL3,int zRR1,int zRR2,int zRR3,int zRL1,int zRL2,int zRL3) {
  // loads PWM target values into the target array based on servo angles
  // if the specified angle == 0 then no action is taken
  if (zFR1 > 0) {setServoTgt(zFR1,SFR1);}
  if (zFR2 > 0) {setServoTgt(zFR2,SFR2);}
  if (zFR3 > 0) {setServoTgt(zFR3,SFR3);}
  
  if (zFL1 > 0) {setServoTgt(zFL1,SFL1);}
  if (zFL2 > 0) {setServoTgt(zFL2,SFL2);}
  if (zFL3 > 0) {setServoTgt(zFL3,SFL3);}

  if (zRR1 > 0) {setServoTgt(zRR1,SRR1);}
  if (zRR2 > 0) {setServoTgt(zRR2,SRR2);}
  if (zRR3 > 0) {setServoTgt(zRR3,SRR3);}
  
  if (zRL1 > 0) {setServoTgt(zRL1,SRL1);}
  if (zRL2 > 0) {setServoTgt(zRL2,SRL2);}
  if (zRL3 > 0) {setServoTgt(zRL3,SRL3);}

  // we also set the start positions to save doing this elsewhere
  SetStartsToPPs();
}

// ---------------------------------------------------------------------------------

void MoveToAllAtRest() {
  // set up moves to retun all servos to rest SQUAT position
  clearMem(); // prime memory loader

  // S:  FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(160, 90, 95,160, 90, 85,160, 90, 95,160, 90, 85,430);

  HeadTask = 0; LEDTaskPnt = -1;
  Move_As_Memory(0,0,50); // row,mode,speed
  AtRest = true; // set droid at rest flag
//  Serial.println("MoveToRest");
}

// ---------------------------------------------------------------------------------

void MoveToBalanceSteps() {
  // set up moves to raise and lower opposite legs continuously
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(130, 70, 95,130, 70, 85,130, 70, 95,130, 70, 85,420);
  MemSet(110, 65, 95,130, 70, 85,130, 70, 95,110, 65, 85,405);
  MemSet( -1,  3, 10,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 10
  MemSet(130, 70, 95,130, 70, 85,130, 70, 95,130, 70, 85,410);
  MemSet( -1,  3, 30,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 30
  MemSet(130, 70, 95,110, 65, 85,110, 65, 95,130, 70, 85,405);
  MemSet( -1,  3, 10,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 10
  MemSet(130, 70, 95,130, 70, 85,130, 70, 95,130, 70, 85,410);
  MemSet( -1,  3, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 15
  MemSet( -1,  4,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Flag = 1
  MemSet( -1,  3, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 15
  MemSet( -1,  1,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Goto row 1

  HeadTask = 0; LEDTaskPnt = -1;
  Move_As_Memory(0,1,50); // row,mode,speed
//  Serial.println("MoveToBalanceSteps");
}

// ---------------------------------------------------------------------------------

void MoveToReady(byte zM) {
  // bend both legs into the ready posture
  // zM = 0 only move legs
  // zM = 1 set head in random mode
  clearMem(); // prime memory loader

  // S:  FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(140, 76, 95,140, 76, 85,140, 76, 95,140, 76, 85,442);  // Ready start pos
  MemSet(140, 76, 95,140, 76, 85,140, 76, 95,140, 76, 85,  5);  // settle

  AtRest = false;
  if (zM > 0) {
    WiFiRestCnt = WiFiRestDef; HeadTask = 10;
    eyeB = 10; eyeG = 0; eyeR = 0; LEDTaskPnt = -2;
  }
  Move_As_Memory(0,0,50); // row,mode,speed
  clearMoveFlags(); Ready = true; // set droid ready flag
  Trans = !Trans; // toggle the transpose flag so that movement starts differently each time
  Serial.println("MoveToReady");
}

// ---------------------------------------------------------------------------------

void MoveToReadyStop() {
  // cxalled when droid is moving and the joystick is released
  MoveToReady(0); loopWhileWalk();
}

// ---------------------------------------------------------------------------------

void MoveToRndHeadPos(int zSpeed) {
  // determine a random head position and set things up  
    getHeadAngle(); // get the current head angle
    if (HeadAngle < 90) {
      HeadAngle = 95 + (5 * random(9));  // set an angle between 95 to 135
    } else {
      HeadAngle = 85 - (5 * random(9));  // set an angle between 45 to 85
    }
    SetStartsToPPs();
    setHeadTgt(HeadAngle); HeadSteps = abs(STG[SH0] - SPP[SH0]); HeadStep = HeadSteps;
    setHeadSpeed(zSpeed); Head = 1;
}

// ---------------------------------------------------------------------------------

void MoveToRndRates() {
  // perform a seemingly random move sequence
  // IMPORTANT - robot must be on its back to avoid damage!!!
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(  0,  0, 95,  0,  0, 85,  0,  0, 95,  0,  0, 85,430);
  MemSet(  0,  0, 95,  0,  0, 85,  0,  0, 95,  0,  0, 85,404);
  MemSet( 90, 90, 95,412,418, 85,446,426, 95, 90, 90, 85,440);
  MemSet( 90, 90, 95,413,420, 85,451,428, 95, 90, 90, 85,404);
  MemSet(  0,  0, 95,423,434, 85,485,448, 95,  0,  0, 85,430);
  MemSet(  0,  0, 95,424,436, 85,490,450, 95,  0,  0, 85,404);
  MemSet(160, 90, 95,431,445, 85,513,463, 95,160, 90, 85,420);
  MemSet(160, 90, 95,437,454, 85,536,476, 95,160, 90, 85,420);
  MemSet(  0,  0, 95,443,463, 85,160, 90, 95,  0,  0, 85,420);
  MemSet(  0,  0, 95,447,468, 85,160, 90, 95,  0,  0, 85,410);
  MemSet( 90, 90, 95,456,482, 85,560,473, 95, 90, 90, 85,430);
  MemSet( 90, 90, 95,459,487, 85,560,467, 95, 90, 90, 85,410);
  MemSet(  0,  0, 95,469,501, 85,560,450, 95,  0,  0, 85,430);
  MemSet(  0,  0, 95,470,502, 85,560,447, 95,  0,  0, 85,404);
  MemSet(160, 90, 95,476,512, 85,560,436, 95,160, 90, 85,420);
  MemSet(160, 90, 95,483,521, 85,560,424, 95,160, 90, 85,420);
  MemSet(  0,  0, 95, 96,140, 85,160,  0, 95,  0,  0, 85,440);
  MemSet(  0,  0, 95, 96,140, 85,160,  0, 95,  0,  0, 85,404);
  MemSet( 90, 90, 95,487,527, 85,539,  0, 95, 90, 90, 85,430);
  MemSet( 90, 90, 95,484,523, 85,532,  0, 95, 90, 90, 85,410);
  MemSet(  0,  0, 95,475,509, 85,511,  0, 95,  0,  0, 85,430);
  MemSet(  0,  0, 95,474,507, 85,509,  0, 95,  0,  0, 85,404);
  MemSet(160, 90, 95,464,494, 85,488,  0, 95,160, 90, 85,430);
  MemSet(160, 90, 95,458,485, 85,473,  0, 95,160, 90, 85,420);
  MemSet(  0,  0, 95,449,471, 85,452,  0, 95,  0,  0, 85,430);
  MemSet(  0,  0, 95,448,470, 85,450,  0, 95,  0,  0, 85,404);
  MemSet( 90, 90, 95,439,456, 85,429,  0, 95,100, 90, 85,430);
  MemSet( 90, 90, 95,436,452, 85,422,  0, 95,100, 90, 85,410);
  MemSet(  0,  0, 95,426,438, 85,  0,  0, 95,  0,  0, 85,430);
  MemSet(  0,  0, 95,425,436, 85,  0,  0, 95,  0,  0, 85,404);
  MemSet(160, 90, 95,416,423, 85,  0,  0, 95,160, 90, 85,430);
  MemSet(160, 90, 95,410,414, 85,  0,  0, 95,160, 90, 85,420);
  MemSet( -1,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Goto row 0

  Move_As_Memory(0,1,50); // row,mode,speed
  Serial.println("MoveToRndRates");
}

// ---------------------------------------------------------------------------------

void MoveToSitHigh() {
  // move to high sitting position
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(160, 46, 95,160, 46, 85, 37,  0, 95, 39,  0, 85,670);  // sit high move
  MemSet(160, 46, 95,160, 46, 85, 37,  0, 95, 39,  0, 85,  5);  // settle
  MemSet(360,246,295,360,246,285,237,200,295,239,200,285,  5);  // pwr off

  Move_As_Memory(0,0,50); // row,mode,speed
  clearMoveFlags(); SitHigh = true; // set droid low sitting flag
  Serial.println("MoveToSitHigh");
}

// ---------------------------------------------------------------------------------

void MoveToSitHigh_Clap() {
  // move to high sitting position from stand, then clap, return to squat
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(130, 70, 95,130, 70, 85,130, 70, 95,130, 70, 85,230);
  MemSet(145, 58, 95,145, 58, 85, 84, 35, 95, 84, 35, 85, 30);
  MemSet(160, 45, 95,160, 45, 85, 37,  0, 95, 37,  0, 85,460);
  MemSet( -1,  3, 50,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 50
  MemSet(160, 39, 95,160, 39, 85, 67, 21, 95, 67, 21, 85,635);
  MemSet(109, 39, 95,109, 39, 85, 67, 19, 95, 67, 19, 85,430);
  MemSet( -1,  3, 50,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 50
  MemSet( -1,  6,  4,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // For...1 to ...4 loop
  MemSet( 92, 39, 77, 92, 39,103, 67, 19, 95, 67, 19, 85,208);
  MemSet( -1,  3,  5,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 5
  MemSet(109, 39, 95,109, 39, 85, 67, 19, 95, 67, 19, 85,410);
  MemSet( -1,  3,  2,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 2
  MemSet( -1,  7,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Next after a For... loop
  MemSet( -1,  3, 50,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 50
  MemSet(160, 39, 95,160, 39, 85, 67, 19, 95, 67, 19, 85,430);
  MemSet(160, 45, 95,160, 45, 85, 37,  0, 95, 37,  0, 85,460);
  MemSet( 82, 71, 95, 82, 71, 85, 15,  0, 95, 15,  0, 85,480);
  MemSet( -1,  3, 75,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 75
  MemSet(  0,  0, 95,  0,  0, 85,  0,  0, 95,  0,  0, 85,660);
  MemSet(200,200,295,200,200,285,200,200,295,200,200,285,610);

  Move_As_Memory(0,0,50); // row,mode,speed
  Serial.println("MoveToSitHigh_Clap");
}

// ---------------------------------------------------------------------------------

void MoveToSitHighAndClap() {
  // move to high sitting position from Ready, then clap, then sit and squat
  Serial.println("MoveToSitHighAndClap");
  HeadTask = 10;                  // move head randomly
  eyeB = 10; LEDTaskPnt = 2;      // blue eyes on
  LEDMthTask = 20; MoveState = 1; // set mouth LEDs to forward motion
  MoveToSitHigh_Clap();
  loopWhileWalk();  // perform move
  clearMoveFlags(); AtRest = true; // set dropid at rest flag
  MoveState = 99;   // wait for joystick release
  LEDMthTask = -1;  // clear the mouth LEDs
}

// ---------------------------------------------------------------------------------

void MoveToSitHighFromLow() {
  // move to high sitting position from low sitting position
  Serial.println("MoveToSitHighFromLow");
  HeadTask = 0;                   // centre the head
  eyeB = 10; LEDTaskPnt = 2;      // blue eyes on
  LEDMthTask = 20; MoveState = 1; // set mouth LEDs to forward motion
  MoveToSitHigh();
  loopWhileWalk();  // perform move
  HeadTask = 10;    // move head randomly
  MoveState = 99;   // wait for joystick release
  LEDMthTask = -1;  // clear the mouth LEDs
}

// ---------------------------------------------------------------------------------

void MoveToSitHighPaw() {
  // move to high sitting position and offer a paw
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(160, 46, 95,160, 46, 85, 37,  0, 95, 39,  0, 85,670);
  MemSet( -1,  3, 75,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 75
  MemSet(107, 25, 95,160, 46, 85, 39,  0, 95, 39,  0, 85,214);
  MemSet( 98, 87, 95,160, 46, 85, 39,  0, 95, 39,  0, 85,455);
  MemSet(298, 87, 95,160, 46, 85, 39,  0, 95, 39,  0, 85,  2);
  MemSet( -1,  3,200,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 200
  MemSet(160, 46, 95,160, 46, 85, 41,  0, 95, 39,  0, 85,455);
  MemSet( -1,  3,100,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 100
  MemSet(160, 46, 95,160, 46, 85, 37,  0, 95, 37,  0, 85,610);
  MemSet(360,246,295,360,246,285, 37,200,295,239,200,285,  5);

  Move_As_Memory(0,0,50); // row,mode,speed
  Serial.println("MoveToSitHigh");
}

// ---------------------------------------------------------------------------------

void MoveToSitLow() {
  // move to low sitting position
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet( 81, 67, 95, 81, 67, 85, 16,  0, 95, 16,  0, 85,670);
  MemSet( 81, 67, 95, 81, 67, 85, 16,  0, 95, 16,  0, 85,  5);
  MemSet(281,267, 95,281,267, 85,216,200, 95,216,200, 85,  5);

  Move_As_Memory(0,0,50); // row,mode,speed
  clearMoveFlags(); SitLow = true; // set dropid low sitting flag
  Serial.println("MoveToSitLow");
}

// ---------------------------------------------------------------------------------

void MoveToSitLowFromRest() {
  // move to low sitting position from Rest
  Serial.println("MoveToSitLowFromRest");
  HeadTask = 0;                   // centre the head
  eyeB = 10; LEDTaskPnt = 2;      // blue eyes on
  LEDMthTask = 20; MoveState = 1; // set mouth LEDs to forward motion
  MoveToSitLow();
  loopWhileWalk();  // perform move
  HeadTask = 10;    // move head randomly
  MoveState = 99;   // wait for joystick release
  LEDMthTask = -1;  // clear the mouth LEDs
}

// ---------------------------------------------------------------------------------

void MoveToSquat() {
  // move to squat position, normal when critical power level is reached
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(  0,  0, 95,  0,  0, 85,  0,  0, 95,  0,  0, 85,705);
  MemSet(  0,  0, 95,  0,  0, 85,  0,  0, 95,  0,  0, 85,  5);
  MemSet(200,200,295,200,200,285,200,200,295,200,200,285,  5);

  Move_As_Memory(0,0,50); // row,mode,speed
  clearMoveFlags(); AtRest = true; // set droid AtRest flag
  Serial.println("MoveToSquat");
}

// ---------------------------------------------------------------------------------

void MoveToSquatAndExercise() {
  // move to squat position, exnormal when critical power level is reached
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(160, 90, 95,160, 90, 85,160, 90, 95,160, 90, 85,230);
  MemSet(113, 74, 95,113, 74, 85,113, 74, 95,113, 74, 85, 30);
  MemSet( 39, 74, 95, 39, 74, 85,113, 74, 95,113, 74, 85,660);
  MemSet( 54, 61, 95, 54, 61, 85, 54, 61, 95, 54, 61, 85,660);
  MemSet( 32, 36, 95, 32, 36, 85, 32, 36, 95, 32, 36, 85,430);
  MemSet( -1,  6,  2,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // For...1 to ...2 loop
  MemSet(116,120, 95,116,120, 85,116,120, 95,116,120, 85,440);
  MemSet(116,120, 95,116,120, 85,116,120, 95,116,120, 85, 20);
  MemSet( 32, 36, 95, 32, 36, 85, 32, 36, 95, 32, 36, 85,440);
  MemSet( 32, 36, 95, 32, 36, 85, 32, 36, 95, 32, 36, 85,420);
  MemSet( -1,  7,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Next after a For... loop
  MemSet( -1,  3, 50,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 50
  MemSet( -1,  6,  3,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // For...1 to ...3 loop
  MemSet( 89, 35, 95, 89, 35, 85, 38, 20, 95, 38, 20, 85,416);
  MemSet( -1,  3, 18,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 18
  MemSet( 32, 36, 95, 32, 36, 85, 32, 36, 95, 32, 36, 85,430);
  MemSet( -1,  3, 20,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 20
  MemSet( -1,  7,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Next after a For... loop
  MemSet( 32, 36,130, 32, 36, 50, 32, 36, 95, 32, 36, 85,220);
  MemSet( 90, 80,130, 90, 80, 50, 32, 36, 95, 32, 36, 85, 15);
  MemSet(153,140,130,153,140, 50,  3,  0, 95,  3,  0, 85,415);
  MemSet( -1,  3,200,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 200
  MemSet(104, 80,130,104, 80, 50, 32, 36, 95, 32, 36, 85,260);
  MemSet( 32, 36,130, 32, 36, 50, 32, 36, 95, 32, 36, 85, 60);
  MemSet( 32, 36, 95, 32, 36, 85, 32, 36, 95, 32, 36, 85,450);
  MemSet(  0,  0, 95,  0,  0, 85,  0,  0, 95,  0,  0, 85,624);

  Move_As_Memory(0,0,50); // row,mode,speed
  clearMoveFlags(); AtRest = true; // set droid AtRest flag
  Serial.println("MoveToSquatAndExercise");
}

// ---------------------------------------------------------------------------------

void MoveToSquatExercise() {
  // move from standing to squat and exercise
  Serial.println("MoveToSquatExercise");
  HeadTask = 10;                  // move head randomly
  eyeB = 10; LEDTaskPnt = 2;      // blue eyes on
  LEDMthTask = 20; MoveState = 1; // set mouth LEDs to forward motion
  MoveToSquatAndExercise();
  loopWhileWalk();  // perform move
  clearMoveFlags(); AtRest = true; // set dropid at rest flag
  MoveState = 99;   // wait for joystick release
  LEDMthTask = -1;  // clear the mouth LEDs
}

// ---------------------------------------------------------------------------------

void MoveToStand() {
  // move to standing position from a squat position
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet( 30,  0, 95, 30,  0, 85, 30,  0, 95, 30,  0, 85,228);
  MemSet(160, 90, 95,160, 90, 85,160, 90, 95,160, 90, 85,484);
  MemSet(160, 90, 95,160, 90, 85,160, 90, 95,160, 90, 85,  5);
  MemSet(360,290, 95,360,290, 85,360,290, 95,360,290, 85,  5);


  Move_As_Memory(0,0,50); // row,mode,speed
  clearMoveFlags(); Stand = true; // set dropid Stand flag
  Serial.println("MoveToStand");
}

// ---------------------------------------------------------------------------------

void MoveToStandFromAny() {
  // move to standing position from any position
  Serial.println("MoveToStandFromAny");
  HeadTask = 0;                   // centre the head
  eyeB = 10; LEDTaskPnt = 2;      // blue eyes on
  LEDMthTask = 20; MoveState = 1; // set mouth LEDs to forward motion
  if (AtRest) {MoveToStand();}
  else if (Ready) {MoveToStandFromReady();}
  else if (SitHigh) {MoveToStandFromSitHigh();}
  loopWhileWalk();  // perform move
  HeadTask = 10;    // move head randomly
  MoveState = 99;   // wait for joystick release
  LEDMthTask = -1;  // clear the mouth LEDs
}

// ---------------------------------------------------------------------------------

void MoveToStandFromReady() {
  // move to the Stand position from the Ready position
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(140, 76, 95,140, 76, 85,140, 76, 95,140, 76, 85, 10);
  MemSet(160, 90, 95,160, 90, 85,160, 90, 95,160, 90, 85,653);
  MemSet(160, 90, 95,160, 90, 85,160, 90, 95,160, 90, 85,  5);
  MemSet(360,290, 95,360,290, 85,360,290, 95,360,290, 85,  5);

  Move_As_Memory(0,0,50); // row,mode,speed
  clearMoveFlags(); Stand = true; // set droid Stand flag
  Serial.println("MoveToStandFromReady");
}

// ---------------------------------------------------------------------------------

void MoveToStandFromSitHigh() {
  // move to standing position from a high sitting position
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(117, 73, 95,117, 73, 85, 48, 34, 95, 48, 34, 85,670);
  MemSet( 60, 34, 95, 60, 34, 85, 68, 34, 95, 68, 34, 85,670);
  MemSet(160, 90, 95,160, 90, 85,160, 90, 95,160, 90, 85,705);
  MemSet(360,290, 95,360,290, 85,360,290, 95,360,290, 85,  5);

  Move_As_Memory(0,0,50); // row,mode,speed
  clearMoveFlags(); Stand = true; // set droid sitting high flag
  Serial.println("MoveToStandFromSitHigh");
}

// ---------------------------------------------------------------------------------

void MoveToStandFromWalk() {
  // move to standing position from walking
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(160, 90, 95,160, 90, 85,160, 90, 95,160, 90, 85,705);
  MemSet(160, 90, 95,160, 90, 85,160, 90, 95,160, 90, 85,  5);
  MemSet(360,290, 95,360,290, 85,360,290, 95,360,290, 85,  5);

  Move_As_Memory(0,0,50); // row,mode,speed
  Serial.println("MoveToStandFromWalk");
}

// ---------------------------------------------------------------------------------

void MoveToStepLeft() {
  // from ready position step to the left
  clearMem(); // prime memory loader
  Trans = false;

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(130, 70, 95,130, 70, 85,130, 70, 95,130, 70, 85,428);
  MemSet(130, 70, 89,130, 70, 79,130, 70, 89,130, 70, 79,630);
  MemSet(130, 70, 89,120, 65, 79,130, 70, 89,130, 70, 79,205);
  MemSet(130, 70, 90, 90, 65,475,130, 70, 90,130, 70, 79,  7);
  MemSet(130, 70, 90, 90, 65, 70,130, 70, 90,130, 70, 79,  7);
  MemSet(130, 70, 90,138, 73, 70,130, 70, 90,127, 69, 79,414);
  MemSet( -1,  3,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 1
  MemSet(130, 70, 90,138, 73, 70,130, 70, 90,120, 65, 79,205);
  MemSet(130, 70, 90,138, 73, 70,130, 70, 90, 90, 65,475,  7);
  MemSet(130, 70, 90,138, 73, 70,130, 70, 90, 90, 65, 70,  7);
  MemSet(130, 70, 90,138, 73, 70,130, 70, 90,138, 73, 70,414);
  MemSet( -1,  3,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 1
  MemSet(138, 73,110,130, 70, 90,138, 73,110,130, 70, 90,630);
  MemSet(120, 65,110,130, 70, 90,138, 73,110,130, 70, 90,205);
  MemSet( 90, 65,505,130, 70, 90,138, 73,110,130, 70, 90,  7);
  MemSet( 90, 65,100,130, 70, 90,138, 73,110,130, 70, 90,  7);
  MemSet(130, 70,100,130, 70, 90,138, 73,110,130, 70, 90,414);
  MemSet( -1,  3,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 1
  MemSet(130, 70,100,130, 70, 90,120, 65,110,130, 70, 90,205);
  MemSet(130, 70,100,130, 70, 90, 90, 65,505,130, 70, 90,  7);
  MemSet(130, 70,100,130, 70, 90, 90, 65,100,130, 70, 90,  7);
  MemSet(130, 70,100,130, 70, 90,130, 70,100,130, 70, 90,414);
  MemSet( -1,  3,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 1
  MemSet( -1,  1,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Goto row 1

  Move_As_Memory(0,1,50); // row,mode,speed
  Serial.println("MoveToStepLeft");
}

// ---------------------------------------------------------------------------------

void MoveToStepRight() {
  // from ready position step to the right
  clearMem(); // prime memory loader
  Trans = false;
  
  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(130, 70, 95,130, 70, 85,130, 70, 95,130, 70, 85,428);
  MemSet(130, 70,101,130, 70, 91,130, 70,101,130, 70, 91,630);
  MemSet(120, 65,101,130, 70, 91,130, 70,101,130, 70, 91,205);
  MemSet( 90, 65,505,130, 70, 90,130, 70,101,130, 70, 90,  7);
  MemSet( 90, 65,110,130, 70, 90,130, 70,101,130, 70, 90,  7);
  MemSet(138, 73,110,130, 70, 90,127, 69,101,130, 70, 90,414);
  MemSet( -1,  3,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 1
  MemSet(138, 73,110,130, 70, 90,120, 65,101,130, 70, 90,205);
  MemSet(138, 73,110,130, 70, 90, 90, 65,505,130, 70, 90,  7);
  MemSet(138, 73,110,130, 70, 90, 90, 65,110,130, 70, 90,  7);
  MemSet(138, 73,110,130, 70, 90,138, 73,110,130, 70, 90,414);
  MemSet( -1,  3,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 1
  MemSet(130, 70, 90,138, 73, 70,130, 70, 90,138, 73, 70,630);
  MemSet(130, 70, 90,120, 65, 70,130, 70, 90,138, 73, 70,205);
  MemSet(130, 70, 90, 90, 65,475,130, 70, 90,138, 73, 70,  7);
  MemSet(130, 70, 90, 90, 65, 80,130, 70, 90,138, 73, 70,  7);
  MemSet(130, 70, 90,130, 70, 80,130, 70, 90,138, 73, 70,414);
  MemSet( -1,  3,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 1
  MemSet(130, 70, 90,130, 70, 80,130, 70, 90,120, 65, 70,205);
  MemSet(130, 70, 90,130, 70, 80,130, 70, 90, 90, 65,475,  7);
  MemSet(130, 70, 90,130, 70, 80,130, 70, 90, 90, 65, 80,  7);
  MemSet(130, 70, 90,130, 70, 80,130, 70, 90,130, 70, 80,414);
  MemSet( -1,  3,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 1
  MemSet( -1,  1,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Goto row 1

  Move_As_Memory(0,1,50); // row,mode,speed
  Serial.println("MoveToStepRight");
}

// ---------------------------------------------------------------------------------

void MoveToTurnLeft() {
  // perform a left neutral turn
  clearMem(); // prime memory loader
  Trans = false;

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(140, 76, 95,140, 76, 85,140, 76, 95,140, 76, 85,428);
  MemSet(143, 76,102,140, 76, 92,143, 76,102,140, 76, 92,428);
  MemSet(113, 76,101,140, 76, 91,140, 76,101,140, 76, 91,207);
  MemSet( 95, 81,494,140, 76,489,140, 76,499,140, 76,489, 10);
  MemSet(143, 81, 82,132, 76, 84,140, 76, 94,140, 76, 84,414);
  MemSet( -1,  3,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 1
  MemSet(143, 81, 82,113, 76, 84,140, 76, 94,140, 76, 84,207);
  MemSet(143, 81, 82, 95, 81,478,140, 76, 94,140, 76, 84, 10);
  MemSet(143, 81,487,147, 81, 72,140, 76,496,140, 76,486, 10);
  MemSet( -1,  3,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 1
  MemSet(145, 76,494,141,480,480,143, 76,498,140, 76,488,211);
  MemSet(143, 76,498,141,278,486,120, 66,500,140, 76,490,  7);
  MemSet(143, 76,102,140, 76, 92,143, 76,102,140, 76, 92,407);
  MemSet( -1,  1,  2,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Goto row 2

  Move_As_Memory(0,1,50); // row,mode,speed
  Serial.println("MoveToTurnLeft");
}

// ---------------------------------------------------------------------------------

void MoveToTurnRight() {
  // perform a right neutral turn
  clearMem(); // prime memory loader
  Trans = false;

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(140, 76, 95,140, 76, 85,140, 76, 95,140, 76, 85,428);
  MemSet(140, 76, 88,143, 76, 78,140, 76, 88,143, 76, 78,428);
  MemSet(140, 76, 89,113, 76, 79,140, 76, 89,140, 76, 79,207);
  MemSet(140, 76,491, 95, 81,486,140, 76,491,140, 76,481, 10);
  MemSet(132, 76, 96,143, 81, 98,140, 76, 96,140, 76, 86,414);
  MemSet( -1,  3,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 1
  MemSet(113, 76, 96,143, 81, 98,140, 76, 96,140, 76, 86,207);
  MemSet( 95, 81,502,143, 81, 98,140, 76, 96,140, 76, 86, 10);
  MemSet(147, 81,108,143, 81,493,140, 76,494,140, 76,484, 10);
  MemSet( -1,  3,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Pause for 1
  MemSet(141,480,500,145, 76,486,140, 76,492,143, 76,482,211);
  MemSet(141,278,494,143, 76,482,140, 76,490,120, 66,480,  7);
  MemSet(140, 76, 88,143, 76, 78,140, 76, 88,143, 76, 78,407);
  MemSet( -1,  1,  2,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Goto row 2

  Move_As_Memory(0,1,50); // row,mode,speed
  Serial.println("MoveToTurnRight");
}

// ---------------------------------------------------------------------------------

void MoveToWalk() {
  // perform a walking sequence from the Ready position
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(140, 76, 95,140, 76, 85,140, 76, 95,140, 76, 85,228);
  MemSet(140,474,101,140,476, 91,140,471,101,140,475, 91,428);
  MemSet(140,474,101,140,476, 91,125, 70,101,140,475, 91,204);
  MemSet(140,472,101,140,475, 91,113, 90,101,140,475, 91,414);
  MemSet(140,471,101,140,475, 91,151, 90,101,140,474, 91,414);
  MemSet(140,471,101,140,475, 91,151,490,101,140,474, 91,  5);
  MemSet(125, 70,101,140,474, 91,151,490,101,140,474, 91,204);
  MemSet(113, 90,101,140,474, 91,551,488,101,140,473, 91,414);
  MemSet(154, 90,101,140,474, 91,550,487,101,140,473, 91,414);
  MemSet(154,490,101,140,473, 91,549,486,101,140,473, 91,609);
  MemSet(553,487, 95,140,473, 85,548,484, 95,140,472, 85,228);
  MemSet(551,485, 89,140,472, 79,546,481, 89,140,471, 79,428);
  MemSet(550,484, 89,140,472, 79,546,481, 89,125, 70, 79,204);
  MemSet(550,483, 89,140,471, 79,545,480, 89,113, 90, 79,414);
  MemSet(549,482, 89,140,471, 79,544,479, 89,154, 90, 79,414);
  MemSet(548,481, 89,140,471, 79,544,478, 89,154,490, 79,  5);
  MemSet(548,481, 89,125, 70, 79,543,478, 89,554,490, 79,204);
  MemSet(547,480, 89,113, 90, 79,542,477, 89,553,488, 79,414);
  MemSet(546,478, 89,154, 90, 79,542,475, 89,552,487, 79,414);
  MemSet(546,478, 89,154,490, 79,541,475, 89,551,486, 79,609);
  MemSet( -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // ---
  MemSet(544,476, 95,553,487, 85,540,473, 95,549,484, 85,221);
  MemSet(543,474,101,551,485, 91,138,471,101,548,482, 91,421);
  MemSet(542,473,101,551,485, 91,125, 70,101,547,481, 91,204);
  MemSet(542,472,101,550,483, 91,113, 93,101,546,480, 91,414);
  MemSet(541,471,101,549,482, 91,154, 93,101,545,478, 91,414);
  MemSet(140,471,101,549,482, 91,154,493,101,545,478, 91,  1);
  MemSet(125, 70,101,548,482, 91,154,493,101,544,478, 91,204);
  MemSet(113, 93,101,547,480, 91,154,491,101,543,477, 91,414);
  MemSet(154, 93,101,546,479, 91,553,489,101,542,475, 91,414);
  MemSet(154,493,101,546,479, 91,553,489,101,542,475, 91,  1);
  MemSet(154,491, 95,545,476, 85,551,487, 95,540,473, 85,221);
  MemSet(553,488, 89,543,474, 79,549,484, 89,138,471, 79,421);
  MemSet(552,488, 89,543,474, 79,549,484, 89,125, 70, 79,204);
  MemSet(551,486, 89,542,472, 79,547,482, 89,113, 93, 79,414);
  MemSet(550,484, 89,541,471, 79,546,480, 89,154, 93, 79,414);
  MemSet(550,484, 89,140,471, 79,546,480, 89,154,493, 79,  1);
  MemSet(549,484, 89,125, 70, 79,545,480, 89,154,493, 79,204);
  MemSet(548,482, 89,113, 93, 79,544,478, 89,154,491, 79,414);
  MemSet(547,480, 89,154, 93, 79,543,476, 89,553,489, 79,414);
  MemSet(547,480, 89,154,493, 79,543,476, 89,553,489, 79,  1);
  MemSet( -1,  1, 22,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Goto row 22

  Move_As_Memory(0,1,50); // row,mode,speed
  Serial.println("MoveToWalk");
}

// ---------------------------------------------------------------------------------

void MoveToWalkBackwards() {
  // perform a walking sequence from the Ready position
  clearMem(); // prime memory loader

  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp 
  MemSet(130, 70, 95,130, 70, 85,130, 70, 95,130, 70, 85,428);
  MemSet(531,471, 86,130, 70, 76,530,470, 86,528,470, 76,430);
  MemSet(531,471, 86,118, 64, 76,530,470, 86,528,470, 76,207);
  MemSet(531,472, 86, 92, 41, 76,530,471, 86,528,470, 76,  5);
  MemSet(532,472, 86,141, 67, 76,530,471, 86,126, 71, 76,415);
  MemSet( -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // ---
  MemSet(532,473, 86,541,467, 76,530,471, 86,126, 71, 76, 10);
  MemSet(532,473, 86,541,467, 76,530,471, 86,114, 65, 76,207);
  MemSet(532,473, 86,541,467, 76,530,472, 86, 92, 41, 76,  5);
  MemSet(533,474, 86,540,468, 76,530,472, 86,129, 61, 76,421);
  MemSet( -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // ---
  MemSet(534,475, 86,539,468, 76,530,472, 86,529,461, 76, 10);
  MemSet(136, 78,104,537,470, 94,530,474,104,532,466, 94,654);
  MemSet(118, 64,104,537,470, 94,530,474,104,533,466, 94,207);
  MemSet( 92, 41,104,536,470, 94,530,474,104,533,467, 94,  5);
  MemSet(141, 67,104,535,471, 94,131, 75,104,534,468, 94,421);
  MemSet( -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // ---
  MemSet(541,467,104,535,471, 94,128, 73,104,535,469, 94, 10);
  MemSet(540,467,104,535,472, 94,116, 67,104,535,470, 94,207);
  MemSet(540,468,104,534,472, 94, 92, 41,104,536,470, 94,  5);
  MemSet(138, 69,104,133, 73, 94,141, 67,104,537,472, 94,421);
  MemSet( -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // ---
  MemSet(138, 69,104,535,474, 94,540,467,104,538,473, 94, 10);
  MemSet(537,470, 95,541,479, 85,538,468, 95,539,475, 85,224);
  MemSet(535,471, 86,148, 84, 76,535,469, 86,141, 77, 76,424);
  MemSet(535,471, 86,118, 64, 76,535,469, 86,539,477, 76,207);
  MemSet(535,471, 86, 92, 41, 76,534,469, 86,537,476, 76,  5);
  MemSet(533,472, 86,141, 67, 76,532,470, 86,128, 73, 76,421);
  MemSet( -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // ---
  MemSet(132, 73, 86,141,467, 76,130, 71, 86,128, 73, 76, 14);
  MemSet( -1,  1,  7,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0);  // Goto row 7
 
  Move_As_Memory(0,1,50); // row,mode,speed
  Serial.println("MoveToWalkBackwards");
}

// ---------------------------------------------------------------------------------

void MoveTurnLeft() {
  // called from move state machine
  if (MoveState != 7) {
    // just been asked to turn left
    MoveState = 7; LEDMthTask = 20;
    if (AtRest) {
      // at reast so we need to move to ready position
      MoveToReady(1); loopWhileWalk();
    }
    // initiate left turn if ready
    if (Ready) {MoveToTurnLeft();}
    else {MoveState = -1; return;}
  }
  if (JoyD == 7) {
    // demand is still there so keep moving at demanded speed
    getSpeedJXY(JoyV);
    setLegSpeed(Speed);
  } else {
    // -X demand has dropped so return to ready state
    MoveState = -1;
  }
}

// ---------------------------------------------------------------------------------

void MoveTurnRight() {
  // called from move state machine
  if (MoveState != 3) {
    // just been asked to turn right
    MoveState = 3; LEDMthTask = 20;
    if (Stand) {
      // stood up straight so we need to move to ready position
      MoveToReady(1); loopWhileWalk();
    }
    // initiate right turn if ready
    if (Ready) {MoveToTurnRight();}
    else {MoveState = -1; return;}
  }
  if (JoyD == 3) {
    // demand is still there so keep moving at demanded speed
    getSpeedJXY(JoyV);
    setLegSpeed(Speed);
  } else {
    // +X demand has dropped so return to ready state
    MoveState = -1;
  }
}

// ---------------------------------------------------------------------------------

void Report_PID_Vals() {
  // send all PID parameters to the serial port
  PrintTx += "PID=" + String(pid_p_gain) + ",";
  if (PID_i_En) {PrintTx += String(pid_i_gain) + ",";} else {PrintTx += "--.--,";}
  if (PID_d_En) {PrintTx += String(pid_d_gain) + ",";} else {PrintTx += "--.--,";}
  if (PID_s_En) {PrintTx += String(pid_s_gain);} else {PrintTx += "--.--";}
  PrintTx += "\n";

  PrintTx += "LIM=" + String(PitchGyrDb) + ",";
  PrintTx += String(PitchSbMax) + ",";
  PrintTx += String(YawGyrDb) + ",";
  PrintTx += String(YawSbMax) + "\n";
}

// ---------------------------------------------------------------------------------

void Reset_PID_Vals() {
  // sets all PID parameters to their default values
  setPID_Defaults();
  PrintTx += "PID is RESET\n";
  Report_PID_Vals();
}

// ---------------------------------------------------------------------------------

void SelfBalance() {
  // perform self balance and anti-topple functions when SBEn > 0
  // anti-topple takes priority over self balancing when both are enabled
  int zSBY = 0; int zSBP = 0;
  if ((SBEn & 1) == 1) {
    //#############################################################################
    //
    //  Topple detection functions
    //
    //#############################################################################
    if (YawGyr > topAngMin) {
      // anti-topple tipping to the left
      if (YawGyr < topAngMax) {
        // only raiuse a leg if not walking
        if (Walk == 0) {
          zSBY = (int)((YawGyr - topAngMin) * topAngGain);
          // raise leg angle increases twice as fast as standing leg
          SOF[SFL3] = -zSBY; SOF[SRL3] = -zSBY;
          SOF[SFR3] = zSBY + zSBY; SOF[SRR3] = zSBY + zSBY;
        }
        AutoOffTimer = AutoOffTimeDef;  // hold off auto-timer when tilted
      } else {
        // critical point reached so collapse to squat position
        // prior to doing this, include the offsets in the present position so
        // that it collapse from the true present position
        SPP[SFL3] += SOF[SFL3]; SPP[SFR3] += SOF[SFR3];
        SPP[SRL3] += SOF[SRL3]; SPP[SRR3] += SOF[SRR3];
        clearSerOffsets(); setBalanceOFF();
        Quick = true; SetMainTask(98);  // move to squat position quickly
        return;
      }
    } else if (((SBEn & 1) == 1) && (YawGyr < -topAngMin)) {
      // anti-topple tipping to the right
      if (YawGyr > -topAngMax) {
        // only raiuse a leg if not walking
        if (Walk == 0) {
          zSBY = (int)((YawGyr + topAngMin) * topAngGain);
          // raise leg angle increases twice as fast as standing leg
          SOF[SFL3] = zSBY + zSBY; SOF[SRL3] = zSBY + zSBY;
          SOF[SFR3] = -zSBY; SOF[SRR3] = -zSBY;
        }
        AutoOffTimer = AutoOffTimeDef;  // hold off auto-timer when tilted
      } else {
        // critical point reached so collapse to squat position
        // prior to doing this, include the offsets in the present position
        SPP[SFL3] += SOF[SFL3]; SPP[SFR3] += SOF[SFR3];
        SPP[SRL3] += SOF[SRL3]; SPP[SRR3] += SOF[SRR3];
        clearSerOffsets(); setBalanceOFF();
        Quick = true; SetMainTask(98);  // move to squat position quickly
        return;
      }
    }
    else {
      // level is within +/- topAngMin degrees, so good enough
      if (SBY != zSBY) {
        SOF[SFL3] = 0; // zero SFL3 offset
        SOF[SFR3] = 0; // zero SFR3 offset
        SOF[SRL3] = 0; // zero SRL3 offset
        SOF[SRR3] = 0; // zero SRR3 offset
        zSBY = 0;
      }
    }
  }
  if ((SBEn & 2) == 2) {
    //#############################################################################
    //
    //  Self-balance PID functions
    //
    //#############################################################################
    // self-balance is enabled so perform PID functions
    // YawGyr goes -ve when tipping to the right
    // the thresholds include a deadband for stability
    if (PID_p_En) {
      // PID P-gain is enabled
      if (YawGyr > YawGyrSBU) {
        // Yaw gyro has gone above the upper threshold
        SBYErr = YawGyr - YawGyrSBU;  // this is the +ve error value
      }
      else if (YawGyr < YawGyrSBL) {
        // Yaw gyro has gone below the lower threshold
        SBYErr = YawGyr - YawGyrSBL;  // this is the -ve error value
      }
      else {SBYErr = 0.0; SBYInt = 0.0;}  // we are within the Yaw deadband
      zSBY = -(int)(SBYErr * pid_p_gain); // PID P o/p
    } else {zSBY = 0;}  // p-gain disabled
    if (PID_i_En) {
      // PID I-gain is enabled
      SBYInt -= SBYErr * pid_i_gain;      // PID I o/p
      SBYInt  = constrain(SBYInt,-YawSbMax,YawSbMax); // limit the integrator
      zSBY += (int)SBYInt;                // PID o/p = P + I
    } else {SBYInt = 0.0;}
    zSBY = constrain(zSBY,-YawSbMax,YawSbMax);  // limit the Yaw PID output
    // all legs swing left/right in harmony
    SOF[SFL3] = zSBY; SOF[SRL3] = zSBY;
    SOF[SFR3] = zSBY; SOF[SRR3] = zSBY;

    if (PID_p_En) {
      if (PitchGyr > PitchGyrSBU) {
        // Pitch gyro has gone above the upper threshold
        SBPErr = PitchGyr - PitchGyrSBU;  // this is the +ve error value
      }
      else if (PitchGyr < PitchGyrSBL) {
        // Pitch gyro has gone below the lower threshold
        SBPErr = PitchGyr - PitchGyrSBL;  // this is the -ve error value
      }
      else {SBPErr = 0.0; SBPInt = 0.0;}  // we are within the Yaw deadband
      zSBP = -(int)(SBPErr * pid_p_gain); // PID P o/p
    } else {zSBP = 0;}  // p-gain disabled
    if (PID_i_En) {
      SBPInt -= SBPErr * pid_i_gain;      // PID I o/p
      SBPInt  = constrain(SBPInt,-PitchSbMax,PitchSbMax); // limit the integrator
      zSBP += (int)SBPInt;                // PID o/p = P + I
    } else {SBPInt = 0.0;}
    zSBP = constrain(zSBP,-PitchSbMax,PitchSbMax);  // limit the Pitch PID output
    // all legs swing forwards/backwards in harmony
    SOF[SFL2] = -zSBP; SOF[SRL2] = -zSBP;
    SOF[SFR2] = zSBP; SOF[SRR2] = zSBP;
  }
  // if no change in SBY then exit
  if ((SBY == zSBY) && (SBP == zSBP)) {return;}
  
  // apply the offsets if stationary, otherwise they will be applied by Walk process
  if (Walk == 0) {
    // set offsets and flag that servo is powered is ON
    if (SBY != zSBY) {
      I2Cpwm.setPWM(SFL3, 0, SPP[SFL3] + SOF[SFL3]); PwrMask[SFL3] = true;
      I2Cpwm.setPWM(SFR3, 0, SPP[SFR3] + SOF[SFR3]); PwrMask[SFR3] = true;
      I2Cpwm.setPWM(SRL3, 0, SPP[SRL3] + SOF[SRL3]); PwrMask[SRL3] = true;
      I2Cpwm.setPWM(SRR3, 0, SPP[SRR3] + SOF[SRR3]); PwrMask[SRR3] = true;
    }
    
    if (SBP != zSBP) {
      I2Cpwm.setPWM(SFL2, 0, SPP[SFL2] + SOF[SFL2]); PwrMask[SFL2] = true;
      I2Cpwm.setPWM(SRL2, 0, SPP[SRL2] + SOF[SRL2]); PwrMask[SRL2] = true;
      I2Cpwm.setPWM(SFR2, 0, SPP[SFR2] + SOF[SFR2]); PwrMask[SFR2] = true;
      I2Cpwm.setPWM(SRR2, 0, SPP[SRR2] + SOF[SRR2]); PwrMask[SRR2] = true;
    }
    
    // turn ON remaining channels that may be in a low power state
    setPwrON();
    
    AutoOffTimer = AutoOffTimeDef;  // return to squat after a delay if nothing else happens
  } SBY = zSBY;  SBP = zSBP; // track change instate
}

// ---------------------------------------------------------------------------------

void setBalanceOFF() {
  // turn OFF the self-balance function, if it were on and clears the flags
  if (SBEn != 0) {
    SOF[SFL3] = 0; I2Cpwm.setPWM(SFL3, 0, SPP[SFL3]); // zero SFL3 offset
    SOF[SFR3] = 0; I2Cpwm.setPWM(SFR3, 0, SPP[SFR3]); // zero SFR3 offset
    SOF[SRL3] = 0; I2Cpwm.setPWM(SRL3, 0, SPP[SRL3]); // zero SRL3 offset
    SOF[SRR3] = 0; I2Cpwm.setPWM(SRR3, 0, SPP[SRR3]); // zero SRR3 offset
  
    SOF[SFL2] = 0; I2Cpwm.setPWM(SFL2, 0, SPP[SFL2]); // zero SFL3 offset
    SOF[SFR2] = 0; I2Cpwm.setPWM(SFR2, 0, SPP[SFR2]); // zero SFR3 offset
    SOF[SRL2] = 0; I2Cpwm.setPWM(SRL2, 0, SPP[SRL2]); // zero SRL3 offset
    SOF[SRR2] = 0; I2Cpwm.setPWM(SRR2, 0, SPP[SRR2]); // zero SRR3 offset
  }

  SBEn = 0;
  SBY = 0; YawGyrSB0 = 0.0; SBYInt = 0.0; SBYErr = 0.0;
  SBP = 0; PitchGyrSB0 = 0.0; SBPInt = 0.0; SBPErr = 0.0;
}

// ---------------------------------------------------------------------------------

void setBalanceON(int zSBEn) {
  // turn ON the self-balance function, using the current gyro angles as targets
  // self-balance deband offset threshols are also set at this point
  // zSBEn = 1  - topple enabled only
  // zSBEn = 2  - self-balance enabled only
  // zSBEn = 3  - tople and self-balance enabled
  SBEn = zSBEn;
  if (SBEn > 1) {
    SBY = 0; YawGyrSB0 = YawGyr; SBYInt = 0.0;
    YawGyrSBU = YawGyr + YawGyrDb; YawGyrSBL = YawGyr - YawGyrDb;
    
    SBP = 0; PitchGyrSB0 = PitchGyr; SBPInt = 0.0;
    PitchGyrSBU = PitchGyr + PitchGyrDb; PitchGyrSBL = PitchGyr - PitchGyrDb;
  }
}

// ---------------------------------------------------------------------------------

void setHeadSpeed(int zS) {
  // sets the head timer to match a speed of zS steps per second
  // max speed is 250 per second (4ms steps)
  headInterval = 1000000/zS;
  if (headInterval < 4000) {headInterval = 4000;} // limit to 250 steps / sec
  headMicros = micros() + 1000; // set head loop timer to trigger after 1ms
}

// ---------------------------------------------------------------------------------

void setHeadTgt(int zHA) {
  // sets the servo target value to match an angle of zHA°
  // ensure that zHA is within the limits set for this servo
  if (zHA < 30) {zHA = 30;}
  else if (zHA > 150) {zHA = 150;}
  STG[SH0] = map(zHA,45,135,SLL[SH0],SUL[SH0]);
}

// ---------------------------------------------------------------------------------

void setLegSpeed(int zS) {
  // sets the leg timer to match a speed of zS steps per second
  // max speed is 180 degrees per second (5.5ms steps) servos can't achieve this
  // RateProf sets the type of timer profile to use
  // a profile consists of 10 speeds: Prof1[] = {200,155,120, 93, 72, 56, 44, 34, 26, 20}
  // where 20 represents normal speed and 200 is 10 times slower
  if (zS < 1) {zS = 1;} // limit minimum speed setting to 1 step/sec.
  walkInterval = 1000000/zS;  // default walk interval
  
//  if (RateProf > 0) {walkInterval = 50000/zS;}  // will be multiplied by 20 - 200 from profile
  if (walkInterval < 4000) {walkInterval = 4000;} // limit to 250 steps / sec
  walkMicros = micros() + walkInterval; // set walk loop timer to trigger and load the 1st step of data
}

// ---------------------------------------------------------------------------------

void setPitchSB(int zVal) {
  // outputs zVal as a Pitch self-balance offset
  // also ensures that power is applied to all servos
  SBP = zVal;
  SOF[SFL2] = -SBP; I2Cpwm.setPWM(SFL2, 0, SPP[SFL2] + SOF[SFL2]); PwrMask[SFL2] = true; // set SFL2 offset
  SOF[SFR2] =  SBP; I2Cpwm.setPWM(SFR2, 0, SPP[SFR2] + SOF[SFR2]); PwrMask[SFR2] = true; // set SFR2 offset
  SOF[SRL2] = -SBP; I2Cpwm.setPWM(SRL2, 0, SPP[SRL2] + SOF[SRL2]); PwrMask[SRL2] = true; // set SRL2 offset
  SOF[SRR2] =  SBP; I2Cpwm.setPWM(SRR2, 0, SPP[SRR2] + SOF[SRR2]); PwrMask[SRR2] = true; // set SRR2 offset
  setPwrON();
}

// ---------------------------------------------------------------------------------

void setPwrON() {
  // turns power ON for all servos and sets the PwrMasks = true  
  I2Cpwm.setPWM(SFL1,0,SPP[SFL1] + SOF[SFL1]); PwrMask[SFL1] = true;
  I2Cpwm.setPWM(SRL1,0,SPP[SRL1] + SOF[SRL1]); PwrMask[SRL1] = true;
  I2Cpwm.setPWM(SFR1,0,SPP[SFR1] + SOF[SFR1]); PwrMask[SFR1] = true;
  I2Cpwm.setPWM(SRR1,0,SPP[SRR1] + SOF[SRR1]); PwrMask[SRR1] = true;

  I2Cpwm.setPWM(SFL2,0,SPP[SFL2] + SOF[SFL2]); PwrMask[SFL2] = true;
  I2Cpwm.setPWM(SRL2,0,SPP[SRL2] + SOF[SRL2]); PwrMask[SRL2] = true;
  I2Cpwm.setPWM(SFR2,0,SPP[SFR2] + SOF[SFR2]); PwrMask[SFR2] = true;
  I2Cpwm.setPWM(SRR2,0,SPP[SRR2] + SOF[SRR2]); PwrMask[SRR2] = true;

  I2Cpwm.setPWM(SFL3,0,SPP[SFL3] + SOF[SFL3]); PwrMask[SFL3] = true;
  I2Cpwm.setPWM(SRL3,0,SPP[SRL3] + SOF[SRL3]); PwrMask[SRL3] = true;
  I2Cpwm.setPWM(SFR3,0,SPP[SFR3] + SOF[SFR3]); PwrMask[SFR3] = true;
  I2Cpwm.setPWM(SRR3,0,SPP[SRR3] + SOF[SRR3]); PwrMask[SRR3] = true;
}

// ---------------------------------------------------------------------------------

void setServoAng(int zA, int zCh) {
  // sets the servo value to match an angle of zA°
  // ensure that angle zA° is within the limits set for this servo
  // note tier 3 servos have assymetrical limits. ie. 130-70,110-50
  int zAL = 0; int zAU = 160; // common angle limits
  int zSL = 0; int zSU = 160; // common servo calibration angles
  byte zM = 0;  // changes mapping process, 0 = default, 1 = reverse, 2 = no correction

  // define the angle limits for a given servo
  switch(zCh) {
    case SH0: zAL = 45; zAU = 135; break;
    
    case SFL1: zAL =  0; zAU = 160; zSL =  0; zSU = 160; zM = 1; break;
    case SFL2: zAL =  0; zAU = 140; zSL =  0; zSU =  90; zM = 3; break;
    case SFL3: zAL = 50; zAU = 110; zSL = 70; zSU = 110;         break;
    
    case SFR1: zAL =  0; zAU = 160; zSL =  0; zSU = 160;         break;
    case SFR2: zAL =  0; zAU = 140; zSL =  0; zSU =  90; zM = 2; break;
    case SFR3: zAL = 70; zAU = 130; zSL = 70; zSU = 110;         break;
    
    case SRL1: zAL =  0; zAU = 160; zSL =  0; zSU = 160; zM = 1; break;
    case SRL2: zAL =  0; zAU = 140; zSL =  0; zSU =  90; zM = 3; break;
    case SRL3: zAL = 50; zAU = 110; zSL = 70; zSU = 110;         break;
    
    case SRR1: zAL =  0; zAU = 160; zSL =  0; zSU = 160;         break;
    case SRR2: zAL =  0; zAU = 140; zSL =  0; zSU =  90; zM = 2; break;
    case SRR3: zAL = 70; zAU = 130; zSL = 70; zSU = 110;         break;
  }
  if (servoOE) {SetServoOE(0);} // turn the servos ON
  // remove rate adjustment flag if necessary
  if (zA >= 400) {zA -= 400;}
  
  // check that servo angle indicates servo is ON
  if (zA >= 200) {
    // angles > 180 indicate servo is to be turned OFF, so we remove the PWM
    // pulses by setting the off-time = 0 and exit
    I2Cpwm.setPWM(zCh, 0, 0); return;
  }
  // apply limits to angle
  if (zA < zAL) {zA = zAL;} else if (zA > zAU) {zA = zAU;}

  // non-linearity of servo is corrected
  switch(zM) {
    case 0:
      // normal polarity with centre value correction
      if (zA >= 90) {
        zA = map(zA,90,zSU,SRV[zCh],SUL[zCh]);
      } else {
        zA = map(zA,zSL,90,SLL[zCh],SRV[zCh]);
      } break;
    case 1:
      // reverse polarity with centre value correction
      if (zA >= 90) {
        zA = map(zA,90,zSU,SRV[zCh],SLL[zCh]);
      } else {
        zA = map(zA,zSL,90,SUL[zCh],SRV[zCh]);
      } break;
    case 2:
      // normal polarity with no centre value correction
      zA = map(zA,zSL,zSU,SLL[zCh],SUL[zCh]); break;
    case 3:
      // reverse polarity with no centre value correction
      zA = map(zA,zSL,zSU,SUL[zCh],SLL[zCh]); break;
  }

  SPP[zCh] = zA; I2Cpwm.setPWM(zCh, 0, zA);
//  if (SI < 1) {PrintTx += "PWM=" + String(zA) + "\n";}
  if (GR == 0) {PrintTx += "PWM=" + String(zA) + "\n";}
}

// ---------------------------------------------------------------------------------

void setServoTgt(int zA, int zCh) {
  // sets the servo target value to match an angle of zA°
  // ensure that angle zA° is within the limits set for this servo
  int zAL = 0; int zAU = 160; // common angle limits
  int zSL = 0; int zSU = 160; // common servo calibration angles
  byte zM = 0;  // changes polarity of mapping process, 0 = default, 1 = reverse

  // define the angle limits for a given servo
  switch(zCh) {
    case SH0: zAL = 45; zAU = 135; break;
    
    case SFL1: zAL =  0; zAU = 160; zSL =  0; zSU = 160; zM = 1; break;
    case SFL2: zAL =  0; zAU = 140; zSL =  0; zSU =  90; zM = 3; break;
    case SFL3: zAL = 50; zAU = 110; zSL = 70; zSU = 110;         break;
    
    case SFR1: zAL =  0; zAU = 160; zSL =  0; zSU = 160;         break;
    case SFR2: zAL =  0; zAU = 140; zSL =  0; zSU =  90; zM = 2; break;
    case SFR3: zAL = 70; zAU = 130; zSL = 70; zSU = 110;         break;
    
    case SRL1: zAL =  0; zAU = 160; zSL =  0; zSU = 160; zM = 1; break;
    case SRL2: zAL =  0; zAU = 140; zSL =  0; zSU =  90; zM = 3; break;
    case SRL3: zAL = 50; zAU = 110; zSL = 70; zSU = 110;         break;
    
    case SRR1: zAL =  0; zAU = 160; zSL =  0; zSU = 160;         break;
    case SRR2: zAL =  0; zAU = 140; zSL =  0; zSU =  90; zM = 2; break;
    case SRR3: zAL = 70; zAU = 130; zSL = 70; zSU = 110;         break;
  }
  // apply limits to angle
  if (zA < zAL) {zA = zAL;} else if (zA > zAU) {zA = zAU;}

  // map angle to servo PWM values gainded from calibration
  // non-linearity of servo is corrected at mid-point
  switch(zM) {
    case 0:
      // normal polarity with centre value correction
      if (zA >= 90) {
        STG[zCh] = map(zA,90,zSU,SRV[zCh],SUL[zCh]);
      } else {
        STG[zCh] = map(zA,zSL,90,SLL[zCh],SRV[zCh]);
      } break;
    case 1:
      // reverse polarity with centre value correction
      if (zA >= 90) {
        STG[zCh] = map(zA,90,zSU,SRV[zCh],SLL[zCh]);
      } else {
        STG[zCh] = map(zA,zSL,90,SUL[zCh],SRV[zCh]);
      } break;
    case 2:
      // normal polarity with no centre value correction
      STG[zCh] = map(zA,zSL,zSU,SLL[zCh],SUL[zCh]); break;
    case 3:
      // reverse polarity with no centre value correction
      STG[zCh] = map(zA,zSL,zSU,SUL[zCh],SLL[zCh]); break;
  }
}

// ---------------------------------------------------------------------------------

void SetStartsToPPs() {
  // sets all start pointers to present positions
  for (int zP = 0; zP < 16; zP++) {SSP[zP] = SPP[zP];}
}

// ---------------------------------------------------------------------------------

void SetTargetsToPPs() {
  // sets all target pointers to present positions, effectively stopping any movement
  for (int zP = 0; zP < 16; zP++) {STG[zP] = SPP[zP];}
}

// ---------------------------------------------------------------------------------

void SetTargetsToSRVs() {
  // sets all target pointers to default rest values
  for (int zP = 0; zP < 16; zP++) {STG[zP] = SRV[zP];}
}

// ---------------------------------------------------------------------------------

void setYawSB(int zVal) {
  // outputs zVal as a Yaw self-balance offset
  // also ensures that power is applied to all servos
  SBY = zVal;
  SOF[SFL3] = SBY; I2Cpwm.setPWM(SFL3, 0, SPP[SFL3] + SOF[SFL3]); PwrMask[SFL3] = true; // set SFL3 offset
  SOF[SFR3] = SBY; I2Cpwm.setPWM(SFR3, 0, SPP[SFR3] + SOF[SFR3]); PwrMask[SFR3] = true; // set SFR3 offset
  SOF[SRL3] = SBY; I2Cpwm.setPWM(SRL3, 0, SPP[SRL3] + SOF[SRL3]); PwrMask[SRL3] = true; // set SRL3 offset
  SOF[SRR3] = SBY; I2Cpwm.setPWM(SRR3, 0, SPP[SRR3] + SOF[SRR3]); PwrMask[SRR3] = true; // set SRR3 offset
  setPwrON();
}

// ---------------------------------------------------------------------------------

void WaitWhileMove() {
  // called to complete a single move like Rest or Ready
  // do not call for a move which repeats or it will never end
  while ((Head > 0) || (Walk > 0)) {
    // complete squat move
    if (micros() >= walkMicros) {Walk_Engine();}
    if (micros() >= headMicros) {Head_Engine();}
  } synchLoopTimers();
}

// ---------------------------------------------------------------------------------

void Walk_Begin() {
  // called to initiate a walk sequence, setting flags and variables consistently
  SetStartsToPPs();     // reset start positions
  setLegSpeed(Speed);   // set walk speed and rate profile
  Walk1stPnt = true;    // ensures first angles are set on PCA driver
  AtRest = false;       // clear at rest flag
  SitHigh = false;      // clear sitting high flag
  SitLow = false;       // clear sitting low flag
  Stand = false;        // clear standing upright flag
  Walk = 1;             // initiate walking
}

// ---------------------------------------------------------------------------------

void Walk_Engine() {
  // walking uses a separate timer so that the speed of walking can be controlled
  // before walking is enabled ensure that start conditions are set:
  // Walk > 0     - must be 1 or more for the function to run
  // LegStep > 0  - a count down which determines the angle position
  // WalkMem > 0  - points to the next row of data in the memory incremented during load function
  // WalkMode > 1 - if 0 onloy read memory values once, if > 0 reload from start
  walkMicros = micros();  // grab the current timeer
  if (WalkTOF) {
    // timer overflow has occured so wait for micros() timer to overflow too
    if (micros() < lastWalkMicros) {WalkTOF = false;}  // back to normal, clear the flag
    else {return;}  // return to do other things in the mean time
  }
  if (Walk > 0) {
    // the Walk time period is a constant, based on the Speed value
    // the step size is a function of the rate profile
    // 0 = default constant step rate
    // 1 = accelerating rate from slow to fast
    // 2 = decelerating rate from fast to slow
    // 3 = accelerating/decelerating slow to fast to slow
    walkMicros += walkInterval;
    LegStep--;  // reduce the step value towards zero
    if (LegStep > 0) {
      // only do rate profile calcs when target has not yet been reached
      LegStepN = LegStep; // account for LegStep being decremented
      if (RateProf < 1) {
        // this is a linear constant step rate profile
        // there is no need to scale or modify step values
        LegStepsD = LegSteps;
      } else {
        // this is a non-linear profile where step values are taken from an array
        // the array is normalised to a factor of 2000
        if (LegSteps <= 20) {
          // the number of steps is within the range of the look-up table so simply map it 
          // note LegStep = 0 - 19 in this range
          // the array pointer needs to take account of the total number of steps
          // if it is less than 20 to get proportional scaling
          LegStepsD = 2000; LegStepN =  (LegStepN * 19)/(LegSteps - 1);
          switch(RateProf) {
            case 1: // decelerating profile
              LegStepN = Prof1[LegStepN]; break;
            case 2: // accelerating profile
              LegStepN = Prof2[LegStepN]; break;
            case 3: // accel/decelerating profile
              LegStepN = Prof3[LegStepN]; break;
          }
        } else {
          // the number of steps exceed the range of the look-up array so we have to
          // extrapolate the intermediate points
          LegStepsD = LegSteps * 100; // LegStepsD = (LegSteps * 2000)/20
          ProfPnt = (LegStep * 20)/LegSteps;
          long LegStepM = (LegStep * 20) % LegSteps;
          switch(RateProf) {
            case 1: // decelerating profile
              LegStepN = (LegSteps * Prof1[ProfPnt])/20;
              LegStepN += (LegStepM * (Prof1[ProfPnt + 1] - Prof1[ProfPnt]))/20;
              break;
            case 2: // accelerating profile
              LegStepN = (LegSteps * Prof2[ProfPnt])/20;
              LegStepN += (LegStepM * (Prof2[ProfPnt + 1] - Prof2[ProfPnt]))/20;
              break;
            case 3: // accel/decelerating profile
              LegStepN = (LegSteps * Prof2[ProfPnt])/20;
              LegStepN += (LegStepM * (Prof3[ProfPnt + 1] - Prof3[ProfPnt]))/20;
              break;
          }
        }
  //      PrintTx += "ProfPnt " + String(ProfPnt) + "\n";
      }
//      Serial.print("Step " + String(LegStepN) + "," + String(LegStepsD) + "," + String((float)(LegStepN/(float)LegStepsD)) + "\n");
    }
  } else {
    // not walking so set timer to default 20ms interval
    walkMicros += walkIntDef;
  }

  lastWalkMicros = micros(); if (walkMicros < micros()) {WalkTOF = true;} // time point has over-flowed, so set overflow flag
//    LedTick = true; // perform LED sequences at same rate as walking
  if (Pause) {return;}  // pause flag causes movement to be put on hold
  if (Walk > 0) {
    switch (Walk) {
      // This Walk_Engine function moves servo values from their start positions to
      // their target positions, tracked by their present position as they are moved.
      case 1: // moving leg servos from start positions to targets 
//        PrintTx += String(LegStep) + "\n";
        if (servoOE) {SetServoOE(0);} // turn the servos ON
        if (LegStep > 0) {
          // only adjust values that have start points different from their targets
          for (ChMapPnt = 0; ChMapPnt < 12;ChMapPnt++) {
            SerPnt = ChMap[ChMapPnt];
            if (PwrMask[ChMapPnt]) {
              // this servo is not flagged as being OFF
              if (SSP[SerPnt] != STG[SerPnt]) {
                if (RateMask[ChMapPnt]) {
                  // this is a rate value so use linear steps, ignoring any profile
                  SPP[SerPnt] = STG[SerPnt] - ((LegStep * (STG[SerPnt] - SSP[SerPnt]))/LegSteps);
                } else {
                  // this is a normal value, so apply rate profile is set
                  SPP[SerPnt] = STG[SerPnt] - ((LegStepN * (STG[SerPnt] - SSP[SerPnt]))/LegStepsD);
                } I2Cpwm.setPWM(SerPnt, 0, SPP[SerPnt] + SOF[SerPnt]);
              }
              else if (Walk1stPnt) {I2Cpwm.setPWM(SerPnt, 0, SPP[SerPnt] + SOF[SerPnt]);}
              else if (SOF[SerPnt] != 0) {I2Cpwm.setPWM(SerPnt, 0, SPP[SerPnt] + SOF[SerPnt]);}
            } else {
              // this servo is set to OFF by an angle >= 200, so turn off PWM
              I2Cpwm.setPWM(SerPnt, 0, 0);
            }
          } Walk1stPnt = false;
        } else {
          // last step, so move all servos to target positions
          // set all start positions == target positions, for next move
          for (ChMapPnt = 0; ChMapPnt < 12;ChMapPnt++) {
            SerPnt = ChMap[ChMapPnt];
            if (PwrMask[ChMapPnt]) {
              // this servo is not flagged as being OFF
              if (SSP[SerPnt] != STG[SerPnt]) {
                SPP[SerPnt] = STG[SerPnt]; SSP[SerPnt] = STG[SerPnt]; I2Cpwm.setPWM(SerPnt, 0, SPP[SerPnt] + SOF[SerPnt]);
              }
              else if (Walk1stPnt) {I2Cpwm.setPWM(SerPnt, 0, SPP[SerPnt] + SOF[SerPnt]);}
              else if (SOF[SerPnt] != 0) {I2Cpwm.setPWM(SerPnt, 0, SPP[SerPnt] + SOF[SerPnt]);}
            } else {
              // this servo is set to OFF by an angle >= 200, so turn off PWM
              I2Cpwm.setPWM(SerPnt, 0, 0);
            }
          } Walk1stPnt = false; Walk = 0;
          // at end of leg move. Check to see if values are being loaded from Mem[]

          if (WalkMem > 0) {
            // values are being loaded from memory, so load the next row
            if (WalkMode < 1) {
              // single move sequence, only read data up to end of memory
              if (WalkMem < MemMax) {
                if (MoveMemTgtLoader(WalkMem)) {
                  // data loaded ok so carry on
                  setLegSpeed(Speed); // set walk speed
                  Walk = 1; // re-initiate walking with the new targets loaded
                }
              } else {PrintTx += "WalkEnd\n";}
            } else {
              // continuous move sequence, only stop if an error occurs
              if (WalkMem >= MemMax) {WalkMem = Mem1st;} // reset pointer if end of memory reached
              if (MoveMemTgtLoader(WalkMem)) {
                // data loaded ok so carry on
                setLegSpeed(Speed); // set walk speed
                Walk = 1; // re-initiate walking with the new targets loaded
              }
            }
          } else {
            // if WalkMem has been set to zero during walking the process will be
            // terminated
            if (SBEn != 0) {setBalanceOFF();} // turn OFF self-balance if not continuing
          }
        }
//        PrintTx += String(SPP[SR0]) + "," + String(SPP[SL0]) + ",";
//        PrintTx += String(SPP[SR1]) + "," + String(SPP[SL1]) + ",";
//        PrintTx += String(SPP[SR2]) + "," + String(SPP[SL2]) + ",";
//        PrintTx += String(SPP[SR3]) + "," + String(SPP[SL3]) + ",";
//        PrintTx += String(SPP[SR4]) + "," + String[SL4]) + "\n";
        break;
    }
  } else {
    // not walking so reset the walk timer
    if (walkInterval != walkIntDef) {walkInterval = walkIntDef;}
  }
}

// ---------------------------------------------------------------------------------
