// ################################################################################
//
//  RoboDog v0.01
//
//  Released:  06/11/2020
//
//  Author: TechKnowTone
//
// ################################################################################
/*
    TERMS OF USE: This software is furnished "as is", without technical support, and
    with no warranty, expressed or implied, as to its usefulness for any purpose. In
    no event shall the author or copyright holder be liable for any claim, damages,
    or other liability, whether in an action of contract, tort or otherwise, arising
    from, out of or in connection with the software or the use or other dealings in
    the software.

    microcontroller:  ESP8266 NodeMCU

    This code is for the RoboDog project. The following features are included:

    * Response to commands from the Move Trainer
    * 16-ch calibration tool works with this code
    * ESP Now WiFi protocol is included for Wii control
    * MPU6050 accelerometers calibrated
    * MPU6050 gyroscopes self calibrate
    * Displays a spirit level

*/
// Declare Libraries
#include <Arduino.h>
#include <Wire.h>
#include <Adafruit_PWMServoDriver.h>
#include <Adafruit_NeoPixel.h>
#include <ESP8266WiFi.h>        // ESP8266 WiFi library
#include <espnow.h>             // ESPNOW peer-to-peer library
#include "SSD1306Wire.h"        // OLED display library

// create I2C PWM driver instance for PCA9685 board
Adafruit_PWMServoDriver I2Cpwm = Adafruit_PWMServoDriver(0x40);

// define VL53LOX constants
#define ADDRESS_DEFAULT 0b0101001 // I2C address 0x29 HEX
#define calcMacroPeriod(vcsel_period_pclks) ((((uint32_t)2304 * (vcsel_period_pclks)* 1655) + 500) / 1000)
#define checkTimeoutExpired() (io_timeout > 0 && ((uint16_t)millis() - timeout_start_ms) > io_timeout)
#define decodeVcselPeriod(reg_val) (((reg_val) + 1) << 1)
#define encodeVcselPeriod(period_pclks) (((period_pclks) >> 1) - 1)
#define startTimeout() (timeout_start_ms = millis())
#define XSHUT 2  // D2 is connected to the XSHUT pin

// Declare register list from VL53L0X.h API, ordered as listed there
#define SYSRANGE_START 0x00
#define SYSTEM_THRESH_HIGH 0x0C
#define SYSTEM_THRESH_LOW 0x0E
#define SYSTEM_SEQUENCE_CONFIG 0x01
#define SYSTEM_RANGE_CONFIG 0x09
#define SYSTEM_INTERMEASUREMENT_PERIOD 0x04
#define SYSTEM_INTERRUPT_CONFIG_GPIO 0x0A
#define GPIO_HV_MUX_ACTIVE_HIGH 0x84
#define SYSTEM_INTERRUPT_CLEAR 0x0B
#define RESULT_INTERRUPT_STATUS 0x13
#define RESULT_RANGE_STATUS 0x14
#define RESULT_CORE_AMBIENT_WINDOW_EVENTS_RTN 0xBC
#define RESULT_CORE_RANGING_TOTAL_EVENTS_RTN 0xC0
#define RESULT_CORE_AMBIENT_WINDOW_EVENTS_REF 0xD0
#define RESULT_CORE_RANGING_TOTAL_EVENTS_REF 0xD4
#define RESULT_PEAK_SIGNAL_RATE_REF 0xB6
#define ALGO_PART_TO_PART_RANGE_OFFSET_MM 0x28
#define I2C_SLAVE_DEVICE_ADDRESS 0x8A
#define MSRC_CONFIG_CONTROL 0x60
#define PRE_RANGE_CONFIG_MIN_SNR 0x27
#define PRE_RANGE_CONFIG_VALID_PHASE_LOW 0x56
#define PRE_RANGE_CONFIG_VALID_PHASE_HIGH 0x57
#define PRE_RANGE_MIN_COUNT_RATE_RTN_LIMIT 0x64
#define FINAL_RANGE_CONFIG_MIN_SNR 0x67
#define FINAL_RANGE_CONFIG_VALID_PHASE_LOW 0x47
#define FINAL_RANGE_CONFIG_VALID_PHASE_HIGH 0x48
#define FINAL_RANGE_CONFIG_MIN_COUNT_RATE_RTN_LIMIT 0x44
#define PRE_RANGE_CONFIG_SIGMA_THRESH_HI 0x61
#define PRE_RANGE_CONFIG_SIGMA_THRESH_LO 0x62
#define PRE_RANGE_CONFIG_VCSEL_PERIOD 0x50
#define PRE_RANGE_CONFIG_TIMEOUT_MACROP_HI 0x51
#define PRE_RANGE_CONFIG_TIMEOUT_MACROP_LO 0x52
#define SYSTEM_HISTOGRAM_BIN 0x81
#define HISTOGRAM_CONFIG_INITIAL_PHASE_SELECT 0x33
#define HISTOGRAM_CONFIG_READOUT_CTRL 0x55
#define FINAL_RANGE_CONFIG_VCSEL_PERIOD 0x70
#define FINAL_RANGE_CONFIG_TIMEOUT_MACROP_HI 0x71
#define FINAL_RANGE_CONFIG_TIMEOUT_MACROP_LO 0x72
#define CROSSTALK_COMPENSATION_PEAK_RATE_MCPS 0x20
#define MSRC_CONFIG_TIMEOUT_MACROP 0x46
#define SOFT_RESET_GO2_SOFT_RESET_N 0xBF
#define IDENTIFICATION_MODEL_ID 0xC0
#define IDENTIFICATION_REVISION_ID 0xC2
#define OSC_CALIBRATE_VAL 0xF8
#define GLOBAL_CONFIG_VCSEL_WIDTH 0x32
#define GLOBAL_CONFIG_SPAD_ENABLES_REF_0 0xB0
#define GLOBAL_CONFIG_SPAD_ENABLES_REF_1 0xB1
#define GLOBAL_CONFIG_SPAD_ENABLES_REF_2 0xB2
#define GLOBAL_CONFIG_SPAD_ENABLES_REF_3 0xB3
#define GLOBAL_CONFIG_SPAD_ENABLES_REF_4 0xB4
#define GLOBAL_CONFIG_SPAD_ENABLES_REF_5 0xB5
#define GLOBAL_CONFIG_REF_EN_START_SELECT 0xB6
#define DYNAMIC_SPAD_NUM_REQUESTED_REF_SPAD 0x4E
#define DYNAMIC_SPAD_REF_EN_START_OFFSET 0x4F
#define POWER_MANAGEMENT_GO1_POWER_FORCE 0x80
#define VHV_CONFIG_PAD_SCL_SDA__EXTSUP_HV 0x89
#define ALGO_PHASECAL_LIM 0x30
#define ALGO_PHASECAL_CONFIG_TIMEOUT 0x30

// define VL53L0C data structures
struct SequenceStepEnables {boolean tcc, msrc, dss, pre_range, final_range;};
struct SequenceStepTimeouts {
      uint16_t pre_range_vcsel_period_pclks, final_range_vcsel_period_pclks;
      uint16_t msrc_dss_tcc_mclks, pre_range_mclks, final_range_mclks;
      uint32_t msrc_dss_tcc_us, pre_range_us, final_range_us;};

// declare VL53L0X variables
uint8_t address = ADDRESS_DEFAULT;
bool did_timeout;
uint16_t io_timeout;
uint8_t last_status;      // status of last I2C transmission
unsigned long LTOF_t0;    // fps timer
unsigned long LTOF_t1;    // fps timer
int LTOF_fps;             // LTOF fps value
uint32_t measurement_timing_budget_us;
byte Pulse = 0;           // unfiltered Range transient pulse counter
uint16_t Range;           // raw 16-bit range value
int RangeFtd;             // filtered range value used by general code
uint16_t RangeLast;       // 16-bit range value
bool RangeNew;            // set to true when a new Range value is determined
uint8_t Status;           // range status data, indicates quality of range value
uint8_t stop_variable;    // read by init and used when starting measurement; is StopVariable field of VL53L0X_DevData_t structure in API
uint16_t timeout_start_ms;
enum vcselPeriodType {VcselPeriodPreRange, VcselPeriodFinalRange};
void VL53L0X_startContinuous(uint32_t period_ms = 0);
inline void VL53L0X_setTimeout(uint16_t timeout) { io_timeout = timeout; }
bool VL53L0X_timeoutOccurred(void);
bool VL53L0X_getSpadInfo(uint8_t * count, bool * type_is_aperture);

// Define general constants
#define AcAvgDivMax 250       // max averaging divider, 250 == 10 second average
#define AutoOffTimeDef 1000;  // default auto-off time-out
#define BatCal 103.66         // calibrated voltage multiplying factor, default 102.3 
#define BatCritical 683       // critical battery threshold @6.6v, default = 683, set == 0 to ignore
#define BatMax 850            // A0 for fully charged battery voltage, == 8.2v
#define BatPin A0             // analog input pin connected to battery voltage divider
#define BatUSB 519            // if <= then USB supply assumed
#define ColBlu 0x000088;      // pallet colour
#define ColCya 0x004444;      // pallet colour
#define ColGrn 0x008800;      // pallet colour
#define ColPur 0x440044;      // pallet colour
#define ColRed 0x880000;      // pallet colour
#define ColYel 0x444400;      // pallet colour
#define D0  16                // map pin to GPIO
#define D1   5                // map pin to GPIO
#define D2   4                // map pin to GPIO
#define D3   0                // map pin to GPIO
#define D4   2                // map pin to GPIO
#define D5  14                // map pin to GPIO
#define D6  12                // map pin to GPIO
#define D7  13                // map pin to GPIO
#define D8  15                // map pin to GPIO
#define D9   3                // map pin to GPIO
#define D10  1                // map pin to GPIO
#define GyAvgDivMax 250       // max averaging divider, 250 == 10 second average
#define ledPin0 D7            // RGB LED serial pin for body LEDs
#define ledPin1 D8            // RGB LED serial pin for eye LEDs
#define ledNum0 5             // number of mouth RGB LEDs connected
#define ledNum1 2             // number of eye RGB LEDs connected
#define LtofDet 335           // max range for target detection
#define LtofMax 350           // out of range range value
#define MemRows 200           // no. rows in the move memory array
#define MemWidth 13           // no. of bytes in a memory row, servo angles + steps
#define Menu0_Battery   0     // a menu 0 display mapping
#define Menu0_Demos     1     // a menu 0 display mapping
#define Menu0_WiFiStats 2     // a menu 0 display mapping
#define Menu0_SpiritLvl 3     // a menu 0 display mapping
#define Menu0_LTOFRange 4     // a menu 0 display mapping
#define Menu0_DispOff   5     // a menu 0 display mapping
#define Pin_OE D6             // define PCA9685 OE pin as D6
#define PWM_Max 450           // max PWM value
#define PWM_Min 150           // min PWM value
#define ReadyCntDef 125;      // 40ms default Ready counter timeout, returns to stand
#define sw0Pin D3             // left switch SW0 assigned to pin D3
#define sw1Pin D4             // right switch SW1 assigned to pin D4
#define topAngMin 10.0        // angle topple correction comes into effect
#define topAngMax 30.0        // angle topple correction goes into collapse
#define topAngGain 3.5        // gain of the toppling correect function
#define walkIntDef 20000      // default walk interval 20ms (50Hz)
#define WiFiRestDef 500       // go to rest delay

// NeoPixel parameters
// Parameter 1 = number of pixels in strip
// Parameter 2 = pin number (most are valid)
// Parameter 3 = pixel type flags, add together as needed:
//   NEO_RGB     Pixels are wired for RGB bitstream
//   NEO_GRB     Pixels are wired for GRB bitstream
//   NEO_KHZ400  400 KHz bitstream (e.g. FLORA pixels)
//   NEO_KHZ800  800 KHz bitstream (e.g. High Density LED strip)
//Adafruit_NeoPixel strip = Adafruit_NeoPixel(ledNum, ledPin, NEO_GRB + NEO_KHZ800);
Adafruit_NeoPixel strip0 = Adafruit_NeoPixel(ledNum0, ledPin0, NEO_GRB + NEO_KHZ800);
Adafruit_NeoPixel strip1 = Adafruit_NeoPixel(ledNum1, ledPin1, NEO_GRB + NEO_KHZ800);

// Define servo mapping to PCA9685 channel numbers
// Not all channel numbers are included. Refer to calibration documentation.
// Channels 7, 8, and 9 are not connected
#define SH0  6  // head servo
#define SFL1 0  // front left elbow
#define SFL2 2  // front left arm
#define SFL3 4  // front left shoulder
#define SFR1 1  // front right elbow
#define SFR2 3  // front right arm
#define SFR3 5  // front right shoulder

#define SRL1 15 // rear left elbow
#define SRL2 13 // rear left arm
#define SRL3 11 // rear left shoulder
#define SRR1 14 // rear right elbow
#define SRR2 12 // rear right arm
#define SRR3 10 // rear right shoulder

// Define PWM freq and servo calibration arrays
#define Freq 50  // MG996R servos run at ~50 Hz updates
// Servo:   { 0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F }
int SLL[] = {125,126,260,120,344,143,229,100,100,100,160,357,118,268,135,101};  // servo lower limit values
int SRV[] = {279,325,463,120,394,189,332,300,300,300,209,404,118,476,340,253};  // reference values used to set a servo
int SUL[] = {489,481,463,332,440,230,424,500,500,500,251,446,333,476,506,472};  // servo upper limit values
// Label:   {FL1 FR1 FL2 FR2 FL3 FR3 SH0 NC  NC  NC  RR3 RL3 RR2 RL2 RR1 RL1}

// Display Settings
const int I2C_DISPLAY_ADDRESS = 0x3c;
const int SDA_PIN = D2;
const int SDC_PIN = D1;

// Initialize the oled display for address 0x3c
 SSD1306Wire display(I2C_DISPLAY_ADDRESS, SDA_PIN, SDC_PIN);


// these are the critical controller settings to achieve self-balance
#define pid_p_gain_ref 50.0         // 88.0 Gain setting for the P-controller (1.0 - 15.0)
#define pid_i_gain_ref 1.2          // 4.0 Gain setting for the I-controller (1.18 - 1.5)
#define pid_d_gain_ref 20.0         // 32.0 Gain setting for the D-controller (10 - 30)
#define pid_s_gain_ref 0.005        // 0.005 Gain setting for PID speed governor (0.05 - 0.001)
//#define pid_s_thresh 0.0            // cut-in threshold for speed governor
#define pid_output_trend_limit 100  // self balancing trend limit. Max angle = pid_output_trend_limit * self_balance_pid_inc
#define PitchGyrDbRef 0.1           // Pitch angle deadband / 2
#define PitchSbMaxRef 10            // Pitch self-balance PID max o/p
#define YawGyrDbRef 0.1             // Yaw angle deadband / 2
#define YawSbMaxRef 20              // Yaw max self-balance PID o/p

// Declare PID controller coeficients
bool PID_d_En;            // PID d-gain enable true=ON, false=OFF
float pid_d_gain;         // Gain setting for the D-controller (0.0 - 10.0)
float pid_p_gain;         // Gain setting for the P-controller (0.0 - 10.0)
bool PID_p_En;            // PID p-gain enable true=ON, false=OFF
bool PID_i_En;            // PID i-gain enable true=ON, false=OFF
float pid_i_gain;         // Gain setting for the I-controller (0.0 - 10.0)
//float pid_d_gain = 0.0;    // 32.0 Gain setting for the D-controller (10 - 40)
int pid_out_max;          // limit PID output to prevent PWM overload
//float pid_sb_dbX = 0.0;     // Self-balancing dead band limit
//float pid_sb_dbY = 0.0;     // Self-balancing dead band limit
bool  PID_s_En;           // PID S-gain enable true=ON, false=OFF
float pid_s_gain = 0.0;   // 0.005 Gain setting for PID speed governor (0.05 - 0.001)

// 3-axis MPU 6050 variables
const int MPU_addr=0x68;      // I2C address of the MPU-6050
long AcAvgDiv;                // dividing factor used in averaging
int16_t AcX;                  // MPU raw X-acceleration value
long AcXAvg;                  // averaged AcX value to reduce noise
int16_t AcXL;                 // previous MPU X-acceleration value
int16_t AcXOff;               // offset to be added to raw X-acceleration value
long AcXSum;                  // averaging accumulator
int16_t AcY;                  // MPU raw Y-acceleration value
long AcYAvg;                  // averaged AcY value to reduce noise
float AcYF;                   // floating point Y acceleration value
int16_t AcYL;                 // previous MPU Y-acceleration value
int16_t AcYOff;               // offset to be added to raw X-acceleration value
bool AcYsgn;                  // positive/negative sign flag
long AcYSum;                  // averaging accumulator
int16_t AcZ;                  // MPU raw Z-acceleration value
long AcZAvg;                  // averaged AcZ value to reduce noise
int16_t AcZL;                 // previous MPU Z-acceleration value
int16_t AcZOff;               // offset to be added to raw X-acceleration value
long AcZSum;                  // averaging accumulator
long GyAvgDiv;                // dividing factor used in averaging
int GyCal;                    // counter used to determine gyro calibration period
unsigned long GyT;            // time of gyro readings in milliseconds
unsigned long GyTD;           // time between successive gyro readings in milliseconds
int16_t GyX;                  // MPU raw X-gyroscope value
long GyXAvg;                  // averaged GyX value to reduce noise
int16_t GyXL;                 // previous MPU X-gyroscope value
int16_t GyXOff;               // offset to be added to raw X-gyroscope value
int16_t GyXOffCnt;            // offset counter for gyro centering
long GyXSum;                  // averaging accumulator
int16_t GyY;                  // MPU raw Y-gyroscope value
long GyYAvg;                  // averaged GyY value to reduce noise
int16_t GyYL;                 // previous MPU Y-gyroscope value
int16_t GyYOff;               // offset to be added to raw Y-gyroscope value
int16_t GyYOffCnt;            // offset counter for gyro centering
long GyYSum;                  // averaging accumulator
int16_t GyZ;                  // MPU raw Z-gyroscope value
long GyZAvg;                  // averaged GyZ value to reduce noise
int16_t GyZL;                 // previous MPU Z-gyroscope value
int16_t GyZOff;               // offset to be added to raw Z-gyroscope value
long GyZSum;                  // averaging accumulator


// Wii Transciever MAC: 50:02:91:68:F7:3F
uint8_t broadcastAddress[] = {0x50, 0x02, 0x91, 0x68, 0xF7, 0x3F};

// 250 byte max for Tx/Rx data
// It must match the receivers data structure
typedef struct ESPNOW_Buff250 {
    char ESPdata[250];
} ESPNOW_Buff250;

// Create an 8Byte_message called NunchukDataTx to hold I2C data message
ESPNOW_Buff250 Tx_Buff;

// Create an 8Byte_message to hold incoming sensor readings
ESPNOW_Buff250 Rx_Buff;



// Declare and initialise global variables
int Angle;                    // temp angle variable used in angle calculations
bool AngPwr;                  // used during drawing, true if the servo is powered, false if not
bool AngRate;                 // used during drawing, true if the angle is an auto-rate angle
int Any;                      // any temp value
long AnyLong;                 // any temp value
String Any$;                  // temporary string
bool AtRest;                  // = true when servos driven to rest position
int AutoOffTimer;             // timer used to auto-power off servos
int BatAvg;                   // preload averaging function with high value
int BatAvgMax;                // maximum average value in this period
byte BatCnt;                  // pointer to array values and critical timer
int BatCritCnt;               // timer used in baterycritical testing
int BatData[60];              // battery averages at 1 second intervals
int BatDel;                   // set a 10 second delay in 'life' prediction
int BatDP;                    // battery data pointer, 0 - 59
int BatLast;                  // previous battery average reading
bool BatPwr;                  // true if battery supply detected, false if USB assumed
int BatQ;                     // a measure of battery quality 1 - 5
unsigned long BatRep;         // time when to report battery reading over serial link
int BatSum;                   // preload averaging sum with 20x max value
unsigned long BatT0;          // time of first battery measurement in life prediction forecast
String BatTime$;              // time left displayed
unsigned long BatTms;         // time when battery reading was taken
int BatV0;                    // reference voltage used in prediction forecast
int BatV1;                    // lowest voltage used in prediction forecast
int BatVol;                   // instantaneous battery voltage
int C_Cnt;                    // counter used in 'C' button detection
byte C_Dn;                    // counter for 'C' button down period
byte C_Up;                    // counter for 'C' button up period
// Order:     SFR1,SFR2,SFR3,SFL1,SFL2,SFL3,SRR1,SRR2,SRR3,SRL1,SRL2,SRL3
int ChMap[] = {  1,   3,   5,   0,   2,   4,  14,  12,  10,  15,  13,  11}; // row point to servo channel map
int ChMapPnt;                 // variable used in array priming
int CmdFlag;                  // global flag used when monitoring Walk actions
int CmdForNum;                // number of times a For... Next loop will run
int CmdForRow;                // points to 1st row in a For...Next loop
char cmdMode;                 // command mode
int cmdSgn;                   // cmdVal sign, 0 or -1
char cmdType;                 // command mode type
int cmdVal;                   // value associated with a cmdType
uint8_t colBlu[] = {0,0,0,0,0,0,0,0,0,0}; // mouth colour array - blue
uint8_t colB;                 // mouth colour flag - blue
uint8_t colGrn[] = {0,0,0,0,0,0,0,0,0,0}; // mouth colour array - green
uint8_t colG;                 // mouth colour flag - green
uint8_t colRed[] = {0,0,0,0,0,0,0,0,0,0}; // mouth colour array - red
uint8_t colR;                 // mouth colour flag - red
int CZ;                       // received button values
bool CZ_Wait;                 // flag used for button release wait function
int Dev_ID;                   // remote device ID, 000 for RoboDog
bool Detect;                  // target detection flag; true detected
int DetectCnt;                // target detection counter
bool DispClr;                 // set == true to clear the memory contents
int DispCnt;                  // counter used in display updating
int DispDel;                  // >0 display delay counter, stops updates over I2C
bool DispDisp;                // set = true to immediately display memory contents
bool DispHold;                // set == true to stop display cycling
bool DispNOP;                 // set == true to stop display I2C updates
int DispNxt;                  // display switch timer, default = 50, 2 sec
byte DispPnt;                 // points at the current display mode
byte DispCX,DispCY;           // display cursor coordinates
bool DispWalk;                // set == true when Walk Engine is active
bool Echo;                    // if true send print messages
bool ESC0;                    // SW0 ESC exit current Main Task flag
bool ESC1;                    // SW1 ESC exit current Main Task flag
bool ESConSW;                 // break loop on button switch pressed flag
uint8_t eyeB;                 // eye colour flag - blue
uint8_t eyeG;                 // eye colour flag - green
uint8_t eyeR;                 // eye colour flag - red
byte Gear;                    // gear for speed range, default = 1
int GR;                       // graph reporting flag, if > 0 then inhibit other reporting
int Head;                     // if > 0 then head is moving
int HeadAngle;                // any angle assoicated with head movement
int HdAngMin,HdAngMax;        // lower and upper limits in head scanning
int HeadCnt;                  // counter used in head tasks
int HeadInc;                  // angle increment used in head movement
unsigned long headInterval;   // main loop head movement interval in microseconds
unsigned long headMicros;     // main loop head moving task interval timer in microseconds
long HeadStep;                // the number of steps left in a head move
long HeadSteps;               // the total number of steps in a head move
int HeadTask;                 // head task pointer
bool HeadTOF;                 // head timer overflow flag
byte I2CError;                // used to detect I2C transmission errors
int JoyD;                     // direction value 0 - 8
int JoyD1;                    // direction 1 counter
int JoyV;                     // joystick vector  = SQRT(SQ(JX) + SQ(JY))
int JoyX;                     // received value
int JoyXC;                    // received value offset
int JoyY;                     // received value
int JoyYC;                    // received value offset
char keyChar;                 // any keyboard character
int keyVal;                   // any keyboard value
unsigned long lastHeadMicros; // timer record used in overflow
unsigned long lastWalkMicros; // timer record used in overflow
int LEDCnt;                   // timer counter used for RGB LEDs
bool LEDMask;                 // flag used to inhibit LED tasks
int LEDMthCnt;                // general counter for mouth LEDs
int LEDMthNext;               // next task pointer for mouth LEDs
int LEDMthPnt;                // general pointer for mouth LEDs
int LEDMthTask;               // task pointer for mouth LEDs
byte LEDMthVal;               // values used in LED tasks
int LEDNextTask;              // task to return to after a wait period
bool LED_Nop;                 // == true to block eye LED tasks, default = false
byte LEDPhase;                // LED show phase
bool LEDshow0;                // LED show flag for mouth LEDs
bool LEDshow1;                // LED show flag for eye LEDs
byte LEDSubTask;              // LED subtask pointer, 0 = default
int LEDTaskPnt;               // LED task pointer, 0 = no task
int LEDTaskWait;              // task wait counter
long LegStep;                 // the number of steps left in a leg move
long LegStepN;                // numerator - modified number of steps left in a leg move
long LegSteps;                // the total number of steps in a leg move
long LegStepsD;               // denominator - modified total number of steps in a leg move
byte LP[] = {0,0,0,0,0};      // general pointers for LED mapping
int LTOF_On;                  // > 0 when laser ranging is enabled
unsigned long LTOF_Wait;      // delay used to prevent wasted time in LTOF readings
bool mainEn = true;           // when == false, used as a main task blocking flag
int mainHedCnt;               // any MainTask head movement counter
byte mainMode;                // used to control specific tasks
int mainSubCnt;               // any MainTask subtask counter
int mainSubNext;              // the next subtask performed after a wait
int mainSubTask;              // any MainTask subtask pointer
int mainSubWait;              // any subtask wait value
int MainTask;                 // main task pointer
int MainTaskNext;             // next main task pointer
int MainTaskWait;             // main task wait timer in 10ms steps
int Mem1st;                   // pointer to 1st memory row to be played by Walk engine
int Mem[MemRows * MemWidth];  // memory array used to develop sequences, 10 bytes are used for each move position
byte MemMax;                  // number of rows loaded into memory
byte MemPnt;                  // a pointer used in loading the Mem[] array
int MemRow;                   // Mem[] row pointer used in loading the array
int Menu0;                    // top level 0 menu pointer
int Menu1;                    // level 1 menu pointer
int MenuLvl;                  // menu level pointer, default 0
int mouthCnt;                 // counter used in mouth LED tasks
int moveCycleCnt;             // counter the number of repeat cycles in movements
bool MovePause;               // flag used to introduce pauses in moves
int MoveState;                // move state machine pointer
bool MTM_Step;                // set = true if single stepping moves
String myMAC;                 // MAC address if this device
unsigned long next10ms;       // 10ms loop timer
unsigned long next20ms;       // 20ms loop timer
unsigned long next40ms;       // 40ms loop timer
int overRun;                  // counter used to prevent undemanded movement
bool Pause;                   // if moving pause if true, continue if false
int Ping = 0;                 // Rx 'ping' counter
float PitchAcc;               // pitch angle in degrees base on accelerometer
float PitchGyr;               // pitch angle in degrees base on gyro + acc
float PitchGyrDb;             // Pitch angle deadband / 2
float PitchGyrSBL;            // pitch gyro self-balance lower threshold
float PitchGyrSB0;            // pitch gyro values used as reference in self-balancing mode
float PitchGyrSBU;            // pitch gyro self-balance Upper threshold
int PitchSbMax;               // Pitch self-balance PID max o/p
unsigned long print40ms;      // 40ms variable print timer
int PrintTgt;                 // 0 == serial port, 1 == WiFi ESP-NOW
String PrintTx = "";          // printed strings
int Prof0[] = {   0, 100, 200, 300, 400, 500, 600, 700, 800, 900,1000,1100,1200,1300,1400,1500,1600,1700,1800,1900,2000}; // linear profile
int Prof1[] = {   0, 124, 260, 410, 568, 720, 880,1030,1176,1312,1432,1540,1642,1730,1800,1868,1914,1952,1980,1994,2000}; // accelerating period
int Prof2[] = {   0,   6,  20,  50,  90, 140, 200, 280, 360, 460, 580, 700, 840, 990,1144,1300,1460,1612,1740,1890,2000}; // decelerating period
int Prof3[] = {   0,  10,  44, 100, 190, 296, 430, 580, 740, 886,1000,1136,1292,1446,1594,1722,1826,1900,1956,1988,2000}; // acc/decelerating period
int ProfPnt;                  // pointer to the Profn[] arrays
float pwmFreq = Freq;         // MG996R servos run at ~50 Hz updates
byte PWM_Phase;               // PWM ramp phase 1 = up, 0 = down
bool PwrMask[16];             // power enable mask used by the walk function
bool RateMask[16];            // 'Rate' enable mask used by the walk function
byte RateProf;                // speed rate profile, 0 = constant
bool Quick;                   // set == true to speed up normal move functions
bool readWiiCall;             // flag prevents call stack overflow
bool Ready;                   // set == true when robot is in the ready position
int ReadyCnt;                 // counter used in returning to default Stand position
bool Reboot;                  // set true when reboot is in progress to stop all tasks
int Rnd;                      // any random integer
int RotCnt;                   // auto zero delay counter for Z-axis
float RotDif;                 // rotation angle difference in degrees based on gyro + acc
float RotGyr;                 // rotation angle in degrees based on gyro + acc
int RPnn;                     // reporting values reference in X-Debug mode
int RPT;                      // if > 0 send text labels for graph data
int RTxTimeout;               // counter used to auto-reset if WiFi fails
int Rx_Chk;                   // Rx checksum
int RxCZ;                     // buffered received CZ value
int RxJoyX;                   // buffered received JoyX value
int RxJoyY;                   // buffered received JoyY value
int Rx_len;                   // length of WiFi received data block
int Rx_Pnt;                   // data pointer user by Rx state machine
bool RxRec;                   // set true when a valid frame of data is received
int RxState;                  // receiver state machine state
int Rx_Task;                  // task pointer user by Rx state machine
int RxVal;                    // value received from serial Rx
byte RxWiFi[6];               // array holding 6 bytes of Wii received data
int SBEn;                     // >0 if self balancing is enabled, 1 = topple, 2 = balance, 3 = both
int SBP;                      // self-balance Pitch servo offset
float SBPErr;                 // self-balance Pitch angle error
float SBPInt;                 // self-balance Pitch PID integrator
int SBY;                      // self-balance Yaw servo offset
float SBYErr;                 // self-balance Yaw angle error
float SBYInt;                 // self-balance Yaw PID integrator
byte servoOE;                 // servo board Output Enable; default = 1  OFF
byte servoTgt;                // serevo target channel for future angle data.
int SI;                       // flag used in blanket servo loading
bool SitHigh;                 // true if in the high sitting position
bool SitLow;                  // true if in the low sitting position
int SL1Av;                    // average angle used in steering
int SL2Av;                    // average angle used in steering
int SL3Av;                    // average angle used in steering
int SerPnt;                   // any servo pointer
int SpdMax;                   // maximum speed when moving
int SpdMid;                   // middle speed when moving
int SpdMin;                   // minimum speed when moving
int SpdTop;                   // current top speed when moving
int Speed;                    // default speed value
int SpeedCnt;                 // used to progressively increase speed on full demand
int SpeedTgt;                 // target speed figure, prevents sudden changes
long SOF[16];                 // array of servo offsets, for balance
long SPP[16];                 // array of servo movement present positions
long SSP[16];                 // array of servo movement start positions
bool Stand;                   // true if the dog is standing upright
long Steps;                   // max number of steps to perform in a move
long STG[16];                 // array of servo movement targets
int sw0Cnt;                   // button switch counter
int sw0DwnTime;               // button pressed down time
bool sw0LastState;            // previous state of button switch, HIGH/LOW
bool SW0_Nop;                 // == true to block SW0 read tasks, default = false
bool sw0State;                // state of read button switch pin
int sw0Timer;                 // timer used to detemine button sequences
int sw1Cnt;                   // button switch counter
bool sw1LastState;            // previous state of button switch, HIGH/LOW
bool SW1_Nop;                 // == true to block SW1 read tasks, default = false
bool sw1State;                // state of read button switch pin
int sw1Timer;                 // timer used to detemine button sequences
unsigned long t0,t1;          // millisecond code timers, report when t1 > t0
int Task10ms;                 // 10ms subtask timer
int Task20ms;                 // 20ms subtask timer
int Task40ms;                 // 40ms subtask timer
int tgtCnt;                   // counter used in automatic servo move to target actions
bool Track;                   // false = retreat, true = track mode
bool Trans;                   // if true transpose angles placed in memory
int Tx_PingCnt;               // 1 second Tx ping timer
bool Once;                    // set == true on reset for first button press
bool UpRht;                   // set == true when the robot is the correct way up
bool USBPwr;                  // true if POST detects USB power voltage
int Walk;                     // if > 0 then walk in one of four directions
bool Walk1stPnt;              // ensures first angles are set on PCA driver
long WalkCnt;                 // a counter which records the walking process
long WalkInc;                 // counter increment value
unsigned long walkInterval;   // main loop walking interval in microseconds
int WalkLast;                 // previousl Walk value if any
long WalkLftDrive;            // drive factor for left-hand side, 128 = max, 0 = min
byte WalkMax;                 // speed factor 4 - 12
int WalkMem;                  // pointer to next move row in Mem[] to be read
byte WalkMode;                // set the mode when moving using Mem[] targets
unsigned long walkMicros;     // main loop walking task interval timer in microseconds
int WalkNext;                 // used to store the next walking mode
long WalkRgtDrive;            // drive factor for right-hand side, 128 = max, 0 = min
int WalkSpeed;                // value used to determine WalkMax
int WalkSpeedTO;              // Time-Out counter for WalkSpeed auto-reduction, counts to zero
bool WalkTOF;                 // walk timer overflow flag
int WalkWait;                 // delay counter used in LTOF mode
int WiFiCntC;                 // C button counter for WiFi enable
int WiFiConCnt;               // counter used in connection retry times
bool WiFiConnected;           // == true if a WiFi connection has been established
bool WiFiEn;                  // true when WiFi is enabled with prolonged 'Z' button activity
int WiFiPing;                 // Rx counter used to maintain connected state
int WiFiRestCnt;              // counter used to goi to rest if no WiFi demands
bool WiFiRx;                  // true when a block of data has been received
long WiFiRxBytes;             // number of bytes received over the WiFi link
int WiFiRxErr;                // number of bytes received over the WiFi link
int WiFiTryCnt;               // WiFi try to connect count on error count
int WiFiTryNum;               // WiFi try to connect total count
bool WiFiTx;                  // true when a block of data has been sent successfully
int WiFiTx_CB;                // == 1 call back is clear to send, == -1 resend
int WiFiTx_Chk;               // Tx checksum
int WiFiTx_len;               // >0 when WiFi Tx buffer contains data to be sent
long WiFiTxBytes;             // number of bytes received over the WiFi link
int WiFiTxErr;                // WiFi data error
int WiFiTxErrCnt;             // WiFi Tx error count, leads to disconnection
int WiiType;                  // device type, 0 == Nunchuk, 1 == Classic
float YawAcc;                 // yaw angle in degrees based on accelerometers
float YawGyr;                 // yaw angle in degrees based on gyro + acc
float YawGyrDb;               // Yaw angle deadband / 2
float YawGyrSBL;              // yaw gyro self-balance lower threshold
float YawGyrSB0;              // yaw gyro values used as reference in self-balancing mode
float YawGyrSBU;              // yaw gyro self-balance upper threshold
int YawSbMax;                 // Yaw max self-balance PID o/p
byte Z_Dn;                    // button pressed state; 0 == UP, >= 1 == Down



// ---------------------------------------------------------------------------------

void setup() {
  // Initialise 16 x 12-bit PWM driver I2C interface immediately to limit any kick from
  // some servos going rogue at power-up
  pinMode(Pin_OE, OUTPUT);
  digitalWrite(Pin_OE, HIGH); // disable all PWM outputs to start with
  Wire.begin();
  I2Cpwm.begin();
  I2Cpwm.setPWMFreq(pwmFreq); // set PCA9685 PWM frequency
  ClearPCAop(); digitalWrite(Pin_OE, LOW);

  pinMode(XSHUT,OUTPUT);
  digitalWrite(XSHUT,LOW);   // VL53L0X reset is active LOW

  pinMode(sw0Pin,INPUT_PULLUP); // pullup resistor for switch SW0
  pinMode(sw1Pin,INPUT_PULLUP); // pullup resistor for switch SW0
  
  setDefaults();
  
  Serial.begin(115200);   // baud rate for Rx/Tx comms

  // Initialise neopixel RGB LEDs
  strip0.begin();
  strip0.fill(0,0,5);  // strip.fill(color, first, count);
  strip0.show();       // Initialize all pixels to 'off'
  strip1.begin();
  strip1.fill(0,0,2);  // strip.fill(color, first, count);
  strip1.show();       // Initialize all pixels to 'off'

  WiFi.mode(WIFI_STA);
  myMAC = WiFi.macAddress();
  WiFi.disconnect();

  runPOST();
  
  // Set ESP-NOW Role as bidirectional
  esp_now_set_self_role(ESP_NOW_ROLE_COMBO);

  // Once ESPNow is successfully Init, we will register for Send CB to
  // get the status of Trasnmitted packet
  esp_now_register_send_cb(OnDataSent);
  
  // Register peer
  esp_now_add_peer(broadcastAddress, ESP_NOW_ROLE_COMBO, 1, NULL, 0);
  
  // Register for a callback function that will be called when data is received
  esp_now_register_recv_cb(OnDataRecv);
  
  synchLoopTimers();    // reset loop timers
}

// ---------------------------------------------------------------------------------

void(* resetFunc) (void) = 0; //declare reset function at address 0

// ---------------------------------------------------------------------------------


void loop() {
  // main code which runs repeatedly as multi-tasking functions
  // continuously read the laser range finder, when enabled
  if (Reboot) {return;} // stop everything as a reset is pending

  if (LTOF_On > 0) LTOF_Read_Tasks();
  if (micros() >= walkMicros) {Walk_Engine();}  // repeated to reduce task timing jitter
  if (micros() >= headMicros) {Head_Engine();}
  if (WiFiConnected) {
    if (WiFiRx) {readWiFiRx();} // buffer contains unread data so deal with it before adding new
  }
  readSerial();               // check serial port

  //###############################################################################
  //
  //  10 ms non-critical subtasks (100Hz)
  //
  //###############################################################################
  // These tasks are run in a consequtive ripple to reduce the overhead on more
  // critical code. They performed in reverse case order and will all complete
  // within the 10ms period
  if (Task10ms > 0) {
    switch(Task10ms) {
      case 1: if (LEDshow0) {strip0.show(); LEDshow0 = false;} break;
      case 2: if (Walk < 1) {AutoPwrOFF();} break;
      case 3: readSW1(); break;
      case 4: readSW0(); break;
      case 5: if (LEDshow1) {strip1.show(); LEDshow1 = false;} break;
      case 6: LED_Task(); break;  // perform head eye LED tasks
    } Task10ms--; // count down the sub-tasks from 6 - 0 as we loop
  }
  else if (millis() >= next10ms) {next10ms += 10; Task10ms = 6;}

  if (micros() >= walkMicros) {Walk_Engine();}  // repeated to reduce task timing jitter

  //###############################################################################
  //
  //  20 ms Main subtasks (50Hz)
  //
  //###############################################################################
  // Every 20ms we check for a main task and do that if one is invoked. Then we
  // perform a series of none critical functions 
  // in a consecutive sequence.
  if (Task20ms > 0) {
    switch(Task20ms) {
      case  1: Head_Tasks(); break;    // head movement tasks
      case  2: MouthTask(); break;     // mouth LED tasks
      case  3: if (SBEn != 0) {SelfBalance();} break;
      case  4: calcAngles(); break;    // calculate yaw, pitch and turning
      case  5: AvgGyrAll(); break;     // average sensor values
      case  6: AvgAccAll(); break;     // average sensor values
      case  7: readAccelAll(); break;  // read MPU6050 accelerometers
      case  8: readGyrosAll(); break;  // read MPU6050 gyroscopes
    } Task20ms--; // count down the sub-tasks from 2 - 0 as we loop
  }
  else if (millis() >= next20ms) {
    next20ms += 20;
    if (mainEn) {Do_Main_Tasks();}  // do main task unless it is blocked
    Task20ms = 8;
  }

  if (micros() >= walkMicros) {Walk_Engine();}  // repeated to reduce task timing jitter

  //###############################################################################
  //
  //  40 ms Tasks (25Hz)
  //
  //###############################################################################
  if (Task40ms > 0) {
    switch(Task40ms) {
      case  1: Display_40ms(); break;  // update display if necessary
      case  2: readBattery(0); break;  // battery monitoring task
      case  3: if (Ping > 0) {Ping--;} break;  // time-out data received flag
      case  4: // if not connected to WiFi run then try to connect every second
        if (!WiFiConnected) {
          // only try to connect WiFi if no 'Pings' are being received
          if (Ping == 0) {
            WiFiConCnt--; if (WiFiConCnt < 1) {WiFiConCnt = 50; WiFiTryToConnect();}
          }
        } else {
          // connceted to WiFi so read Wii data and respond
          readWii();
        } break;
      case  5: if (WiFiPing > 0) {
        WiFiPing--;
        if (WiFiPing < 1) {WiFiDisconnect();} // we have stopped receiving so disconnect WiFi
        } break;
    } Task40ms--;
  }
  else if (millis() >= next40ms) {
    next40ms += 40; Task40ms = 5;
    if (GR > 0) {GR--;} // slowly remove graph reporting inhibit
    Rnd = random(8);    // spin the random wheel for luck
  }

  if (micros() >= walkMicros) {Walk_Engine();}  // repeated to reduce task timing jitter

  //###############################################################################
  //
  //  40 ms Print Task - Variable
  //
  //###############################################################################
  // this code uses either the serial port or the WiFi link to send messages
  if (millis() >= print40ms) {
    print40ms = millis() + 40;
    if (!Echo) {PrintTx = "";}  // Echo is false so empty print buffer
    if (PrintTx.length() > 0) {
      // characters in string buffer so send some/all of them
      print40ms = millis() + 4; // shorten the delay for more data to send
      Tx_PingCnt = 0;           // reset the WiFi ping counter
      if (PrintTgt == 0) {
        // default uses the serial port
        if (Serial.availableForWrite() >= 64) {
          if (PrintTx.length() <= 64) {
            Serial.print(PrintTx);
            PrintTx = "";   // empty the buffer
          } else {
            Serial.print(PrintTx.substring(0,64));
            PrintTx = PrintTx.substring(64);
          }
        }
      } else {
        // text data has been received from WiFi so respond using ESP_NOW
        // ensure that we have had a call back from the previous frame
        if (WiFiConnected) {
          // only trannsmit PrintTx data if connected
          if (WiFiTx_CB == 1) {
            WiFiTx_CB = 0;  // clear the call back flag when sending
            if (PrintTx.length() <= 32) {
              WiFiTx_len = PrintTx.length();
              PrintTx.toCharArray(Tx_Buff.ESPdata,WiFiTx_len + 1);
              WiFiTxBytes += WiFiTx_len;  // track number of bytes sent
              WiFiTxGetChecksum(WiFiTx_len); Tx_Buff.ESPdata[WiFiTx_len] = WiFiTx_Chk; WiFiTx_len++;
//              t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
              PrintTx = "";   // empty the buffer
            } else {
              WiFiTx_len = 32;
              Any$ = PrintTx.substring(0,WiFiTx_len);
              Any$.toCharArray(Tx_Buff.ESPdata,WiFiTx_len + 1);
              WiFiTxBytes += WiFiTx_len;  // track number of bytes sent
              WiFiTxGetChecksum(WiFiTx_len); Tx_Buff.ESPdata[WiFiTx_len] = WiFiTx_Chk; WiFiTx_len++;
//              t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
              PrintTx = PrintTx.substring(32);
            }
          } else if (WiFiTx_CB == -1) {
            // an error has occured so resend the data
            if (WiFiTryCnt < 5) {
              WiFiTryCnt++;   // track the number of retries
              WiFiTx_CB = 0;  // clear the call back flag when sending
//              t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
            } else {
              // too many resends so disconnect WiFi link
              WiFiDisconnect();
              Serial.println("Failed to Tx!");
            }
          }
        } else {
          // not connected so clear PrintTxdata
          PrintTx = "";
        }
      }
    }
  }
  if (t1 > t0) {
    // code debug timer used to check time needed by functions
    PrintTx += String(t1- t0) + "\n"; t0 = t1; // report and reset timers
  }
}

// ---------------------------------------------------------------------------------

void loopDelay(long zDelay) {
  // called to run the main loop for a zDelay period in ms before returning
  // note the calling function must block being called from loop() to prevent stack
  // overflow conditions
  zDelay += millis(); // set the exit time
  while (millis() < zDelay) {
    loop();   // run tasks in the main loop
    yield();  // pass control to background functions
  }
}

// ---------------------------------------------------------------------------------

void loopWhileSwDwn() {
  // called to run the main loop whilst either SW0 or SW1 is pressed down
  // note the calling function must block being called from loop() to prevent stack
  // overflow conditions
  while (!digitalRead(sw0Pin) || !digitalRead(sw1Pin)) {
    loop();   // run tasks in the main loop
    yield();  // pass control to background functions
  }
}

// ---------------------------------------------------------------------------------

void loopWhileWalk() {
  // allows functions to run the main loop whilst being called
  // under certain conditions the ESCape flag will be set to exit this function
  while (Walk > 0) {
    loop();     // run tasks in the main loop
    if (ESConSW) {
      if (!sw0State || !sw1State) {ESConSW = false; break;}
    } yield();  // pass control to background functions
  }  
}

// --------------------------------------------------------------------------------

void OnDataRecv(uint8_t * mac, uint8_t *incomingData, uint8_t len) {
  // Callback when data is received
  // we copy the data into a buffer and set a flag so that the main loop will act
  // to handle it
  while (WiFiRx) {
    // buffer contains unread data so deal with it before adding new
    readWiFiRx();
  }

  memcpy(&Rx_Buff, incomingData, len);

  // incoming data will always be two or more characters, ie. text == $+++
  if (len < 2) {Serial.println("Rx Error - no Rx data"); return;}
  
  if (Rx_len > 0) {Serial.println("Rx Error - buffer in use");} // this should not happen

  // if connected Rx data will have a checksum so check it first
  Rx_len = len - 1; Rx_Pnt = 0;
  RxGetChecksum(Rx_len);  // generate a checksum from Rx data
  if (Rx_Chk != Rx_Buff.ESPdata[len - 1]) {
    Serial.println("Rx Error - checksum");
  }
  // set flags before exiting
  WiFiRxBytes += Rx_len; Rx_Task = 0; WiFiRx = true;
  WiFiPing = 50;    // set link time-out to 2 seconds, Wii Transceiver will keep setting this with Wii data
//    Serial.print("Bytes received: "); Serial.println(len);
}

// --------------------------------------------------------------------------------

void OnDataSent(uint8_t *mac_addr, uint8_t sendStatus) {
  // Callback when data is sent, determines success or failure
//  t1 = micros();
//  Serial.print("Send Status: ");
  if (WiFiConnected) {
    // the link is already active so handle normal success/failure scenarios
    if (sendStatus == 0){
  //    Serial.println("Success " + String(t1-t0));
      WiFiTx = true;    // flag WiFi Tx success
      WiFiTx_CB = 1;    // set flag to indicate that a call back state has occured
      WiFiTxErrCnt = 0; // clear the error flag in case of a glitch
      WiFiPing = 50;    // set link time-out to 2 seconds
    } else {
  //    Serial.println("Fail " + String(t1 - t0));
      WiFiTx = false;   // flag WiFi Tx failure
      WiFiTx_CB = -1;   // set flag to indicate that a resend is required
      WiFiTxErrCnt++;
    }
  } else {
    // not connected so we must have tried to connect and may get an error
    if (sendStatus == 0){
      // successful so enter the connected state
      WiFiConnected = true;
      WiFiPing = 50;    // set link time-out to 2 seconds, Wii Transceiver will keep setting this with Wii data
      WiFiTxErrCnt = 0; // reset error count for this connection
    } else {
      WiFiTxErrCnt++;
    }
    WiFiTx_CB = 1;
  }
}

// ---------------------------------------------------------------------------------

void runPOST() {
  // called during start-up
  // Initialise the display
  Display_Init();
  Display_Intro();

  Serial.println("\n\nRoboDog v0.01");
  readBattery(0);
  readBattery(1);
  if (BatVol <= BatUSB) {Serial.println("\nBatVol = USB Pwr");USBPwr = true;}
  
  Serial.println("LED test");
  NeoPixel_SetAllEyeLEDs(10,10,10); // turn all eye LEDs dim white
  strip1.show(); delay(10);
  NeoPixel_SetAllMthLEDs(10,10,10); // turn all mouth LEDs dim white
  strip0.show();
  
  Serial.println("Enabling LTOF");
  VL53L0X_ON();
  
  initialiseMPU();                  // initialise the accelerometer
  delay(1000);                      // display RGD LEDs white for 1 second
  NeoPixel_SetAllEyeLEDs(0,0,0);    // turn all eye LEDs OFF
  NeoPixel_SetAllMthLEDs(0,0,0);    // turn all mouth LEDs OFF
  strip0.show(); delay(5); strip1.show(); delay(5);
  
  Serial.println("MAC: " + myMAC + "\n");
  // Init ESP-NOW
  if (esp_now_init() != 0) {
    Serial.println("Error initializing ESP-NOW");
  } else {Serial.println("ESP-NOW Activated");}

  AutoOffTimer = 100; // set 1s timeout
  SetMainTask(0);
  Serial.println("POST complete!");
}

// ---------------------------------------------------------------------------------

void setDefaults() {
  // load default values
  AcAvgDiv = 0;           // dividing factor used in averaging
  AcXAvg = 0;             // averaged AcY value to reduce noise
  AcXOff = -38;           // offset to be added to raw X-acceleration value
  AcXSum = 0;             // averaging accumulator
  AcYAvg = 0;             // averaged AcY value to reduce noise
  AcYOff = -14;           // offset to be added to raw X-acceleration value
  AcYsgn = false;         // positive/negative sign flag
  AcYSum = 0;             // averaging accumulator
  AcZAvg = 0;             // averaged AcY value to reduce noise
  AcZOff = 7842 - 8192;   // offset to be added to raw Z-acceleration value
  AcZSum = 0;             // averaging accumulator
  AngPwr = false;         // during drawing true if servo is powered, false if not
  AngRate = false;        // used during drawing, true if the angle is an auto-rate angle
  AtRest = true;          // = true when servos driven to rest position
  AutoOffTimer = 0;       // timer used to auto-power off servos
  BatAvg = 1023;          // preload averaging function with high value
  BatAvgMax = 0;          // maximum average value in this period
  BatCnt = 0;             // pointer to array values and critical timer
  BatCritCnt = 0;         // timer used in baterycritical testing
  BatDel = 250;           // set a 10 second delay in 'life' prediction
  BatDP = 0;              // battery data pointer
  BatLast = 0;            // previous battery average reading
  BatPwr = false;         // true if battery supply detected, false if USB assumed
  BatRep = 0;             // time when to report battery reading over serial link
  BatSum = 20460;         // preload averaging sum with 20x max value
  BatT0 = 0;              // time of first battery measurement in life prediction forecast
  BatTime$ = "--:--";     // time left displayed
  BatV0 = 0;              // reference voltage used in prediction forecast
  BatV1 = 0;              // lowest voltage used in prediction forecast
  BatVol = 0;             // instantaneous battery voltage
  C_Cnt = 0;              // counter used in 'C' button detection
  C_Dn = 0;               // counter for 'C' button down period
  ChMapPnt = 0;           // variable used in array priming
  CmdFlag = 0;            // global flag used when monitoring Walk actions
  CmdForNum = 0;          // number of times a For... Next loop will run
  CmdForRow = 0;          // points to 1st row in a For...Next loop
  cmdMode = ' ';          // command mode
  cmdSgn = 1;             // cmdVal sign, 1 or -1
  cmdType = ' ';          // command mode type
  cmdVal = 0;             // value associated with a cmdType
  colB = 255;             // mouth colour flag - blue
  colG = 255;             // mouth colour flag - green
  colR = 255;             // mouth colour flag - red
  C_Up = 50;              // counter for 'C' button up period
  CZ = 0;                 // received button values
  CZ_Wait = false;        // flag used for button release wait function
  Detect = false;         // target detection flag; true detected
  DetectCnt = 0;          // target detection counter
  Dev_ID = 0;             // remote device ID, 000 for RoboDog
  DispClr = false;        // set == true to clear the memory contents
  DispCnt = 0;            // counter used in display updating
  DispDel = 0;            // >0 display delay counter, stops updates over I2C
  DispDisp = false;       // set = true to immediately display memory contents
  DispHold = false;       // set == true to stop display cycling
  DispNOP = false;        // set == true to stop display I2C updates
  DispNxt = 50;           // display switch timer, default = 50, 2 sec
  DispPnt = 0;            // points at the current display mode
  DispWalk = false;       // set == true when Walk Engine is active
  Echo = true;            // if true send print messages
  ESC0 = false;           // SW0 ESC exit current Main Task flag
  ESC1 = false;           // SW1 ESC exit current Main Task flag
  ESConSW = false;        // break loop on button switch pressed flag
  eyeB = 10;              // eye colour flag - blue
  eyeG = 0;               // eye colour flag - green
  eyeR = 0;               // eye colour flag - red
  Gear = 1;               // gear for speed range, default = 1
  GR = 0;                 // graph reporting flag, if > 0 then inhibit other reporting
  GyAvgDiv = 0;           // dividing factor used in averaging
  GyCal = 250;            // counter used to determine gyro calibration period
  GyX = 0;                // MPU raw X-gyroscope value
  GyXAvg = 0;             // averaged GyX value to reduce noise
  GyXOff = 0;             // offset to be added to raw X-gyroscope value
  GyXOffCnt = 0;          // offset counter for gyro centering
  GyXSum = 0;             // averaging accumulator
  GyY = 0;                // MPU raw Y-gyroscope value
  GyYAvg = 0;             // averaged GyY value to reduce noise
  GyYOff = 0;             // offset to be added to raw Y-gyroscope value
  GyYSum = 0;             // averaging accumulator
  GyZ = 0;                // MPU raw Z-gyroscope value
  GyZAvg = 0;             // averaged GyZ value to reduce noise
  GyZOff = 0;             // offset to be added to raw Z-gyroscope value
  GyYOffCnt = 0;          // offset counter for gyro centering
  GyZSum = 0;             // averaging accumulator
  Head = 0;               // if > 0 then head is moving
  headInterval = 20000;   // main loop head movement interval in microseconds
  HeadTask = 0;           // head task pointer
  HeadTOF = false;        // head timer overflow flag
  I2CError = true;        // used to detect I2C transmission errors
  JoyD1 = 0;              // direction 1 counter
  JoyX = 128;             // received value, default
  JoyY = 128;             // received value, default
  LEDCnt = 0;             // timer counter used for RGB LEDs
  LEDMask = false;        // flag used to inhibit LED tasks
  LEDMthCnt = 0;          // general counter for mouth LEDs
  LEDMthNext = 0;         // next task pointer for mouth LEDs
  LEDMthPnt = 0;          // general pointer for mouth LEDs
  LEDMthTask = 0;         // task pointer for mouth LEDs
  LEDMthVal = 0;          // values used in LED tasks
  LED_Nop = false;        // == true to block eye LED tasks, default = false
  LEDPhase = 0;           // LED show phase
  LEDshow0 = false;       // LED show flag for mouth LEDs
  LEDshow1 = false;       // LED show flag for eye LEDs
  LEDSubTask = 0;         // LED subtask pointer, 0 = default 
  LEDTaskPnt = 2;         // LED task pointer, 0 = no task
  LTOF_On = 0;            // > 0 when laser ranging is enabled, default is 0 == OFF
  MainTask = 0;           // main task pointer
  Mem1st = 0;             // pointer to 1st memory row to be played by Walk engine, default = 0
  MemMax = MemRows;       // number of rows loaded into memory
  MemRow = -1;            // Mem[] row pointer used in loading the array
  Menu0 = 0;              // top level 0 menu pointer
  Menu1 = 0;              // level 1 menu pointer
  MenuLvl = 0;            // menu level pointer, default 0
  mouthCnt = 0;           // counter used in mouth LED tasks
  MovePause = false;      // flag used to introduce pauses in moves
  MoveState = 0;          // move state machine pointer
  MTM_Step = false;       // set = true if single stepping moves
  myMAC = "";             // MAC address if this device
  overRun = 0;            // counter used to prevent undemanded movement
  Pause = false;          // if moving pause if true, continue if false
  Ping = 0;               // 'ping' counter
  PitchAcc = 0.0;         // pitch angle in degrees base on accelerometer
  PitchGyr = 0.0;         // pitch angle in degrees base on gyro + acc
  PrintTx = "";           // clear printed strings
  PrintTgt = 0;           // 0 == serial port, 1 == WiFi ESP-NOW
  ProfPnt = 0;            // pointer to the Profn[] arrays
  Quick = false;          // set == true to speed up normal move functions
  RangeLast = 600;        // 16-bit range value
  readWiiCall = false;    // flag prevents call stack overflow
  RangeNew = false;       // set to true when a new Range value is determined
  RateProf = 0;           // speed rate profile, 0 = constant
  readWiiCall = false;    // flag prevents call stack overflow
  Ready = false;          // set == true when roboty is in the ready position
  ReadyCnt = ReadyCntDef; // counter used in returning to default Stand position
  Reboot = false;         // set true when reboot is in progress to stop all tasks
  RotCnt = 50;            // auto zero delay counter for Z-axis
  RotGyr = 0;             // rotation angle in degrees based on gyro + acc
  RPnn = 0;               // reporting values reference in X-Debug mode
  RPT = 0;                // if > 0 send text labels for graph data
  RTxTimeout = 0;         // counter used to auto-reset if WiFi fails
  Rx_Chk = 0;             // Rx checksum
  RxCZ = 0;               // buffered received CZ value
  RxJoyX = 0;             // buffered received JoyX value
  RxJoyY = 0;             // buffered received JoyY value
  Rx_len = 0;             // length of WiFi received data block
  RxRec = false ;         // set true when a valid frame of data is received
  RxState = 0;            // receiver state machine state
  Rx_Task = 0;            // pointer user by Rx state machine
  RxVal = -1;             // value received from serial Rx
  SBEn = 0;               // >0 if self balancing is enabled, 1 = topple, 2 = balance, 3 = both
  SBP = 0;                // self-balance Pitch servo offset
  SBPInt = 0.0;           // self-balance Pitch PID integrator
  SBY = 0;                // self-balance Yaw servo offset
  SBYInt = 0.0;           // self-balance Yaw PID integrator
  SerPnt = 0;             // any servo pointer
  servoOE = 1;            // servo board Output Enable; default = 1  OFF
  servoTgt = 0;           // serevo target channel for future angle data.
  SI = 0;                 // flag used in blanket servo loading
  SitHigh = false;        // true if in the high sitting position
  SitLow = false;         // true if in the low sitting position
  SpeedCnt = 0;           // used to progressively increase speed on full demand
  Stand = false;          // true if the dog is standing upright
  sw0Cnt = 0;             // button switch counter
  sw0DwnTime = 0;         // button pressed down time
  sw0LastState = HIGH;    // previous state of button switch, HIGH/LOW
  SW0_Nop = false;        // == true to block SW0 read tasks, default = false
  sw0State = HIGH;        // state of read button switch pin
  sw0Timer = 0;           // timer used to detemine button sequences
  sw1Cnt = 0;             // button switch counter
  sw1LastState = HIGH;    // previous state of button switch, HIGH/LOW
  SW1_Nop = false;        // == true to block SW1 read tasks, default = false
  sw1State = HIGH;        // state of read button switch pin
  sw1Timer = 0;           // timer used to detemine button sequences
  t0 = 0;                 // microsecond timer used for code speed checks
  t1 = 0;                 // microsecond timer used for code speed checks
  Task10ms = 0;           // 10ms subtask timer
  Task20ms = 0;           // 20ms subtask timer
  Task40ms = 0;           // 40ms subtask timer
  tgtCnt = 0;             // counter used in automatic servo move to target actions
  Trans = false;          // if true transpose angles placed in memory
  Tx_PingCnt = 0;         // 1 second Tx ping timer
  UpRht = true;           // set == true by AcZ when the robot is the correct way up
  USBPwr = false;         // true if POST detects USB power voltage
  Once = true;            // set == true on reset for first button press
  Walk = 0;               // if >0 then walk in one of four directions
  Walk1stPnt = false;     // ensures first angles are set on PCA driver
  walkInterval = walkIntDef;   // main loop walking interval in microseconds
  WalkLast = 0;           // previousl Walk value if any
  WalkLftDrive = 128;     // drive factor for left-hand side, 128 = max, 0 = min
  WalkMax = 4;            // speed factor 4 - 12
  WalkRgtDrive = 128;     // drive factor for right-hand side, 128 = max, 0 = min
  WalkSpeed = 1;          // value used to determine WalkMax, 1 - 5
  WalkSpeedTO = 0;        // Time-Out counter for WalkSpeed auto-reduction, counts to zero
  WalkTOF = false;        // walk timer overflow flag
  WiFiCntC = 0;           // C button counter for WiFi enable
  WiFiConCnt = 0;         // counter used in connection retry times
  WiFiConnected = false;  // == true if a WiFi connection has been established
  WiFiEn = false;         // true when WiFi is enabled with prolonged 'Z' button activity
  WiFiPing = 0;           // Rx counter used to maintain connected state
  WiFiRx = false;         // true when a block of data has been received
  WiFiRxBytes = 0;        // number of bytes receiv ed over the WiFi link
  WiFiRxErr;              // number of receive errors in the WiFi link
  WiFiTryCnt = 0;         // WiFi try to connect count
  WiFiTryNum = 0;         // WiFi try to connect total count
  WiFiTx = false;         // true when a block of data has been sent successfully
  WiFiTxBytes = 0;        // number of bytes receiv ed over the WiFi link
  WiFiTx_CB = 1;          // == 1 call back is clear to send, == -1 resend
  WiFiTx_Chk = 0;         // Tx checksum
  WiFiTx_len = 0;         // >0 when WiFi Tx buffer contains data to be sent
  WiFiTxErr = 0;          // WiFi data error
  WiFiTxErrCnt = 0;       // WiFi Tx error count, leads to disconnection
  WiiType = 0;            // device type, 0 == Nunchuk, 1 == Classic
  YawAcc = 0.0;           // yaw angle in degrees based on accelerometers
  YawGyr = 0.0;           // yaw angle in degrees based on gyro + acc
  Z_Dn = 0;               // button pressed state; 0 = UP, 1 = Down

  // fill arrays with default values, gained from calibration process
  SPP[ 0] = SUL[ 0];  // SFL1
  SPP[ 1] = SLL[ 1];  // SFR1
  SPP[ 2] = SUL[ 2];  // SFL2
  SPP[ 3] = SLL[ 3];  // SFR2
  SPP[ 4] = SRV[ 4];  // SFL3
  SPP[ 5] = SRV[ 5];  // SFR3
  SPP[ 6] = SRV[ 6];  // SH0
  SPP[ 7] = SRV[ 7];  // NC
  SPP[ 8] = SRV[ 8];  // NC
  SPP[ 9] = SRV[ 9];  // NC
  SPP[10] = SRV[10];  // SRR3
  SPP[11] = SRV[11];  // SRL3
  SPP[12] = SLL[12];  // SRR2
  SPP[13] = SUL[13];  // SRL2
  SPP[14] = SLL[14];  // SRR1
  SPP[15] = SUL[15];  // SRL1

  clearBatData();
  clearSerOffsets();

  setPID_Defaults();
  SetSpeedDefaults();
  SetStartsToPPs();
  SetTargetsToPPs();
  clearMem();
}

// ---------------------------------------------------------------------------------

void setPID_Defaults() {
  // called to load PID default values    
  PID_d_En = false;       // PID D-gain enable true=ON, false=OFF
  pid_d_gain = 0.0;       // Gain setting for the D-controller (0.0 - 10.0)
  pid_p_gain = 0.0;       // Gain setting for the P-controller (0.0 - 10.0)
  PID_p_En = true;        // PID P-gain enable true=ON, false=OFF
  PID_i_En = true;        // PID I-gain enable true=ON, false=OFF
  pid_i_gain = 0.0;       // Gain setting for the I-controller (0.0 - 10.0)
  pid_out_max = 20;       // limit PID output to prevent PWM overload
  PID_s_En = false;       // PID S-gain enable true=ON, false=OFF
  pid_s_gain = 0.0;       // Gain setting for the S-controller (0.0 - 10.0)

  PitchGyrDb = PitchGyrDbRef; // Pitch angle deadband / 2
  PitchGyrSBL = 0.0;      // pitch gyro self-balance lower threshold
  PitchGyrSBU = 0.0;      // pitch gyro self-balance upper threshold
  PitchSbMax = PitchSbMaxRef; // Pitch self-balance PID max o/p

  YawGyrDb = YawGyrDbRef; // Yaw angle deadband / 2
  YawGyrSBL = 0.0;        // yaw gyro self-balance lower threshold
  YawGyrSB0 = 0.0;        // yaw gyro values used as reference in self-balancing mode
  YawGyrSBU = 0.0;        // yaw gyro self-balance upper threshold
  YawSbMax = YawSbMaxRef; // Yaw max self-balance PID o/p
}

// ---------------------------------------------------------------------------------

void synchLoopTimers() {
  // reset main loop timers
  next10ms = millis() + 2;  // reset 10ms loop timer
  next20ms = millis() + 4;  // reset 20ms loop timer with 2ms phase to 10ms timer
  next40ms = millis() + 6;  // reset 40ms loop timer with 4ms phase to 10ms timer
  print40ms = millis() + 8; // reset 40ms printer loop timer with 6ms phase to 10ms timer
}

// --------------------------------------------------------------------------------
